var storage = (function () {

    var db; // wm db
    var collection; // main db collection
    var historyCollection; // history db
    var projectsCollection;
    var stopUpdate = false;


    function __init() {
        openDB();
        setCollection(_settings.data_file_name);
        setAutoSave();
    }


    function openDB() {
        var dbPath = (_settings.data_directory != "") ? path.normalize(_settings.data_directory) : app.dataPath;
        db = new Db(dbPath, {}); // memStore: true , cacheMaxObjSize: 100000
        console.log('Database path: ' + dbPath);
    }

    function closeDB() {
        db.close();
    }

    function setCollection(name) {
        collection = db.collection(name);
        historyCollection = db.collection(name + "_history");
        projectsCollection = db.collection(name + "_projects");

        // create indexes
        collection.createIndex(["dtm", 1], function (err, name) {
            console.log("Index created: " + name)
        });
        collection.createIndex("del", function (err, name) {
            console.log("Index created: " + name)
        });
        collection.createIndex("pid", function (err, name) {
            console.log("Index created: " + name)
        });
        collection.createIndex("fav", function (err, name) {
            console.log("Index created: " + name)
        });
        collection.createIndex("pth", function (err, name) {
            console.log("Index created: " + name)
        });
        collection.createIndex(["ord", -1], function (err, name) {
            console.log("Index created: " + name)
        });
    }





    /*
     ######     #######   ##     ##  ########      ###      ######    ########
    ##    ##   ##     ##  ###   ###  ##     ##    ## ##    ##    ##      ##
    ##         ##     ##  #### ####  ##     ##   ##   ##   ##            ##
    ##         ##     ##  ## ### ##  ########   ##     ##  ##            ##
    ##         ##     ##  ##     ##  ##         #########  ##            ##
    ##    ##   ##     ##  ##     ##  ##         ##     ##  ##    ##      ##
     ######     #######   ##     ##  ##         ##     ##   ######       ##
    */

    // add automatic compact function


    function compactCollection() {
        try {
            var path = app.dataPath + "/" + _settings.data_file_name;
            var oldSize = getFileSizeInMB(path).toString();

            collection.compactCollection(function () {
                closeDB(); // close collection
                __init(); // init again!
                repository.setRepoCollection(); // init repo collection again!
                var newSize = getFileSizeInMB(path).toString();
                wm.showNotice("<strong>Database compacted!</strong>\n\nOld size was: " + oldSize + "MB\nNew size is: " + newSize + " MB", {
                    autoClose: false
                });
                localStorage['DATABASE_LAST_COMPACT'] = new Date().getTime();


            }); // throws error first time

        } catch (e) {
            console.log(e);
        }
    }


    /*

     ########     ###     ######  ##    ## ##     ## ########
     ##     ##   ## ##   ##    ## ##   ##  ##     ## ##     ##
     ##     ##  ##   ##  ##       ##  ##   ##     ## ##     ##
     ########  ##     ## ##       #####    ##     ## ########
     ##     ## ######### ##       ##  ##   ##     ## ##
     ##     ## ##     ## ##    ## ##   ##  ##     ## ##
     ########  ##     ##  ######  ##    ##  #######  ##

    */

    function openSettingFiles() {
        opener(path.normalize("settings/settings_main.json"));
        opener(path.normalize("settings/settings_main_user.json"));
    }

    function openDefaultThemeFile() {
        opener(path.normalize("themes/" + _currentTheme + ".css"));
    }


    function openDBfolder() {
        var dbPath = (_settings.data_directory != "") ? _settings.data_directory : app.dataPath;
        opener(dbPath);
    }


    function openBackupFolder() {
        var basePath = (_settings.backup_directory !== "") ? path.normalize(_settings.backup_directory) :
            (process.platform === 'win32') ? process.env.HOMEPATH : process.env.HOME;
        var targetDirName = '_WRITEMONKEY3_BACKUP';
        var targetDirPath = path.join(basePath, targetDirName);

        // create directory if it doesn't exist
        !fs.existsSync(targetDirPath) && fs.mkdirSync(targetDirPath);
        opener(targetDirPath);
    }


    function backupDatabase(inBackground) {

        var basePath = (_settings.backup_directory !== "") ? path.normalize(_settings.backup_directory) :
            (process.platform === 'win32') ? process.env.HOMEPATH : process.env.HOME;

        var targetDirName = '_WRITEMONKEY3_BACKUP';
        var targetDirPath = path.join(basePath, targetDirName);

        try {
            // create directory if it doesn't exist
            !fs.existsSync(targetDirPath) && fs.mkdirSync(targetDirPath);

            Object.keys(db._cols).forEach(function (key) {
                var source = db._cols[key]._filename;
                var target = path.join(basePath, targetDirName, db._cols[key]._name);
                // rename to older and create new
                fs.rename(target, target + "_OLDER", function (e) { // rename existing file if exists
                    fs.createReadStream(source).pipe(fs.createWriteStream(target)); // then write new one

                    // include custom dic ...
                });
            });

            // set timestamp
            localStorage['DATABASE_LAST_BACKUP'] = new Date().getTime();
            console.log("Database backuped!");



            !inBackground && wm.showNotice('<strong>Database backup completed in:</strong><br><br>' + targetDirPath + '/', {
                autoClose: false
            });

        } catch (e) {
            !inBackground && wm.showNotice('<strong>Error creating database backup in:</strong><br><br>' + targetDirPath + '/', {
                autoClose: false
            });
        }
    }



    /*
    ########   ########    #######         ##
    ##     ##  ##     ##  ##     ##        ##
    ##     ##  ##     ##  ##     ##        ##
    ########   ########   ##     ##        ##
    ##         ##   ##    ##     ##  ##    ##
    ##         ##    ##   ##     ##  ##    ##
    ##         ##     ##   #######    ######
    */



    function newProject(parentId) {
        if (!parentId || _settings.allow_subprojects == false) {
            parentId = "-1";
        }


        var html = 'Enter project name: <br><input autofocus style="font-size: .9em; width: 95%; outline: none; border: none; background-color: transparent;"></input> <div style="font-size: 1em; float: right;">Hit enter to confirm!</div>';


        _editor.openDialog(html, function (value) {
            var ob = getBlankProject();
            ob.nam = value.trim();
            ob.par = parentId;

            projectsCollection.insert(ob, function (err, result) {
                if (!err) {
                    var prId = result[0]._id.valueOf();
                    var prName = result[0].nam;

                    // store empty doc into db
                    var blankO = getBlankDocumentObject();
                    blankO.pid = prId;
                    collection.insert(blankO, function (err, result) {

                        // redraw tree
                        setProjectsTree(function () {
                            // set new project
                            if (parentId != "-1") {
                                $("#sources_list").jqxTree('expandItem', $('#' + parentId)[0]);
                            }
                            setProject(prId);
                        });
                    });
                }
            });

        }, {
            value: "",
            selectValueOnOpen: true,
            closeOnEnter: true,
            closeOnBlur: false,
            onClose: function () {}
        });
    }


    function getBlankProject() {

        var timestamp = new Date().getTime();

        return {
            dtc: timestamp, // date created - only set first time, don't update!!
            dtm: timestamp, // date modified
            par: "-1", // parent id
            nam: "", // project name
            del: false, // is deleted
            arc: false, // is archived
            exc: false, // exclude from compilation
            ord: 0, // order
            wrd: 0 // number of words in project
        };
    }


    function getProjectsArr(callback) {
        projectsCollection
            .find({
                $or: [{
                    del: {
                        $exists: false
                    }
                }, {
                    del: false
                }]
            })
            .sort({
                // par: 1,
                dtm: -1
            })
            .limit(10000)
            .toArray(function (err, arr) {

                // map to ensure par field and sort by par, them dtm
                if (!arr) var arr = [];
                arr = arr.map(function (item) {
                    item.par = (item.par == null || item.par == undefined) ? "-1" : item.par;
                    return item;
                }).sort(function (o1, o2) {
                    return parseInt(o1.par, 10) - parseInt(o2.par, 10) || o2.dtm - o1.dtm;
                });

                callback(err, arr);

            });


    }



    function setProject(prId) {

        _currentProjectId = prId;
        localStorage['PROJECTS_CURRENT_PROJECT_ID'] = prId;
        $("#sources_list").jqxTree('selectItem', $('#' + _currentProjectId.toString())[0]);

        setDocumentList();

    }


    function closeProject() {
        _currentProjectId = -1
        localStorage['PROJECTS_CURRENT_PROJECT_ID'] = "-1";
        $('#sources_list').jqxTree('selectItem', $('#sources_list li')[1]);
    }


    function setProjectsTree(callback) {

        // remove projects from tree
        $('#sources_list').jqxTree('removeItem', $('#pb6')[0]);
        // create new project item
        $('#sources_list').jqxTree('addAfter', {
            "id": "pb6",
            "value": "some value",
            "html": '<strong>Projects</strong>',
            "expanded": true
        }, $('#pb0')[0]);


        // now populate projects
        getProjectsArr(function (err, pArr) {

            var currentProjectsO = {};


            var extendedData = localStorage['PROJECTS_TREE_EXPANDED_DATA'] ? JSON.parse(localStorage['PROJECTS_TREE_EXPANDED_DATA']) : {};


            for (var i = 0; i < pArr.length; i++) {

                var el;
                if (pArr[i].par == -1 || _settings.allow_subprojects == false) {
                    el = $("#sources_list").jqxTree('getItem', $('#pb6')[0]).element;
                } else {
                    // check if parent is not yet set
                    // and move item to the end if it is not
                    // if (currentProjectsO[pArr[i].par] === null) {
                    // 	prArr.push(pArr.splice(i, 1));
                    // 	console.log("correcting for project order");
                    // 	continue;
                    // }
                    el = $("#sources_list").jqxTree('getItem', $('#' + pArr[i].par.toString())[0]).element;
                }

                currentProjectsO[pArr[i]._id.valueOf()] = pArr[i]; // remap project list

                var icon = pArr[i].par == "-1" ? "&#xf07b;" : "&#xf114;";
                // var icon = extendedData[i] && extendedData[i] === true ? "&#xf07b;" : "&#xf114;";



                $('#sources_list').jqxTree('addTo', {
                    id: pArr[i]._id,
                    html: '<span class="awf">' + icon + '</span> ' + pArr[i].nam,
                    value: JSON.stringify(pArr[i]),
                    expanded: extendedData[pArr[i]._id] ? extendedData[pArr[i]._id] : false
                }, el);

            } // END FOR


            _currentProjectsList = currentProjectsO;

            if (callback) callback();

        });
    }


    function showProjectsPalette() {
        getProjectsArr(function (err, arr) {
            var paletteA = [];
            for (var i = 0; i < arr.length; i++) {

                var icon = arr[i].par == "-1" ? "&#xf07b;" : "&#xf114;";
                paletteA.push({
                    id: '',
                    string: arr[i].nam,
                    shortcut: '',
                    action: 'storage.setProject(' + arr[i]._id + ')',
                    icon: icon,
                    type: "eval",
                    sublist: [],
                    persist: false
                    // widget: {
                    // 	type: "eval",
                    // 	value: "";
                    // }
                });
            }

            commandPalette.hideCommandPalette();
            commandPalette.showCommandPalette(paletteA);
            $('.searchFilter').val("");

        });
    }

    function moveProjectToTop(pid) {
        projectsCollection.update({
            _id: pid.toString()
        }, {
            $set: {
                dtm: new Date().getTime()
            }
        }, function (err, numberOfUpdatedDocs) {
            setProjectsTree(function () {
                $('#sources_list').jqxTree('selectItem', $('#' + pid)[0]);
            });
        });
    }

    function renameProject(pid) {

        var html = 'Enter project name: <br><input autofocus style="font-size: .9em; width: 95%; outline: none; border: none; background-color: transparent;"></input> <div style="font-size: 1em; float: right;">Hit enter to confirm!</div>';


        _editor.openDialog(html, function (value) {

            projectsCollection.update({
                _id: pid.toString()
            }, {
                $set: {
                    nam: value
                }
            }, function (err, numberOfUpdatedDocs) {
                if (!err) {
                    var item = $('#' + pid.toString())[0];
                    var icon = _currentProjectsList[pid.toString()].par == "-1" ? "&#xf07b;" : "&#xf114;";

                    item.label = '<span class="awf">' + icon + '</span> ' + value;
                    $('#sources_list').jqxTree('updateItem', $('#' + pid.toString())[0], item);
                    if (_currentProjectsList[pid.toString()]) _currentProjectsList[pid.toString()].nam = value;
                }

            });

        }, {
            value: _currentProjectsList[pid.toString()] ? _currentProjectsList[pid.toString()].nam : "",
            selectValueOnOpen: true,
            closeOnEnter: true,
            closeOnBlur: false,
            onClose: function () {}
        });

    }


    function deleteProject(pid) {

        var modal = new tingle.modal({
            footer: true,
            stickyFooter: false,
            closeMethods: [],
            closeLabel: "",
            beforeClose: function () {
                return true; // close the modal
                //return false; // nothing happens
            }
        });

        modal.setContent('<strong>You are about to delete this project and its subprojects!</strong><br><br>All documents will still be available in <i>Documents</i> view but unsorted.');
        modal.addFooterBtn('Delete project', 'tingle-btn', function () {
            // here goes some logic
            // ids to delete
            var delA = [];
            delA.push(pid.toString()); // add first parent
            findChildren([pid.toString()]); // recursively search for all children

            // search for its children
            function findChildren(idA) {
                var foundChA = [];
                for (var i = 0; i < idA.length; i++) {
                    for (var key in _currentProjectsList) {
                        if (_currentProjectsList[key].par == idA[i]) {
                            var p = _currentProjectsList[key]._id.toString();
                            foundChA.push(p)
                            delA.push(p);
                        }
                    }
                }
                if (foundChA.length > 0)
                    findChildren(foundChA);
            }

            // remove from projects
            projectsCollection.update({
                _id: {
                    $in: delA
                }
            }, {
                $set: {
                    del: true
                }
            }, {
                "multi": true
            }, function (err, numberOfUpdatedDocs) {

                console.log("Number of deleted (sub)projects: ", numberOfUpdatedDocs);

                // select some other tile
                $('#sources_list').jqxTree('selectItem', getNeighborItem($('#' + pid.toString())[0]));
                // remove jq item
                $('#sources_list').jqxTree('removeItem', $('#' + pid.toString())[0]);


                // only removes parent!!!!!!!!!!!!!!!!!!!!!!! but works?????????????
                if (_currentProjectsList[pid.toString()])
                    delete _currentProjectsList[pid.toString()];
            });
            modal.close();
        });

        modal.addFooterBtn('Cancel', 'tingle-btn', function () {
            // here goes some logic
            modal.close();
        });

        // wm theme styling for modal dialogs - tingle
        $(".tingle-btn").css({
            "background": localStorage["VIEWPORT_MARK_COLOR"],
            "color": localStorage["VIEWPORT_BACKGROUND_COLOR"]
        });

        modal.open();
    }


    function getNeighborItem(element) {

        if ($("#sources_list").jqxTree('getNextItem', element)) {
            return $("#sources_list").jqxTree('getNextItem', element);
        } else {
            return $("#sources_list").jqxTree('getPrevItem', element);
        }
    }

    function assignDocumentToProject(event) { // $("#doc_list").on('dragEnd', function)


        // get value of the target via hitTest
        var ev = event.args.originalEvent;
        var x = ev.pageX;
        var y = ev.pageY;
        var item = $('#sources_list').jqxTree('hitTest', x, y);
        var target = JSON.parse(item.value);
        var targetId = target._id;


        // get value of the source
        var source = JSON.parse(event.args.value);
        var sourceId = source._id;
        var sourcePid = source.pid;

        $('li #' + item.element.id).fadeOut(150).fadeIn(70).fadeOut(70).fadeIn(200);


        // update doc with new pid
        collection.update({
            _id: sourceId
        }, {
            $set: {
                pid: targetId,
                ord: new Date().getTime()
            }
        }, function (err, numberOfUpdatedDocs) {

            if (!err) {
                if (_currentProjectId != -1 && targetId != sourcePid) {
                    _currentDoc.pid = targetId;

                    // remove list object if in project
                    $("#doc_list").jqxListBox('unselectIndex', event.args.owner._clickedIndex);
                    $('#doc_list').jqxListBox('removeAt', event.args.owner._clickedIndex);
                    // select next item not working if last on list
                    $("#doc_list").jqxListBox('selectIndex', event.args.owner._clickedIndex);
                }
            }
        });
    }


    function copyProjectTxt(pid) {


        getCombinedProject(pid, function (textsA) {

            var txt = textsA.map(function (item) {
                return stats.removeMetaPart(item.txt).trim();
            }).join('\n\n');

            wm.showNotice('Text from <strong>' + textsA.length.toString() + '</strong> documents copied', {
                autoClose: true
            });


            var clipboard = gui.Clipboard.get();
            clipboard.set(txt, 'text');
        });
    }


    function getCombinedProject(pid, callback) {

        collection
            .find({
                pid: pid,
                del: false,
                rep: false
            }, {
                und: 0
            })
            .sort({
                ord: -1
            })
            .limit(1000)
            .toArray(function (err, docArr) {
                if (!err && docArr.length >= 1) {
                    if (callback) callback(docArr);
                }
            });
    }


    /*
    ##    ##   ########   ##      ##      ########    #######    ######
    ###   ##   ##         ##  ##  ##      ##     ##  ##     ##  ##    ##
    ####  ##   ##         ##  ##  ##      ##     ##  ##     ##  ##
    ## ## ##   ######     ##  ##  ##      ##     ##  ##     ##  ##
    ##  ####   ##         ##  ##  ##      ##     ##  ##     ##  ##
    ##   ###   ##         ##  ##  ##      ##     ##  ##     ##  ##    ##
    ##    ##   ########    ###  ###       ########    #######    ######
    */
    function newDocument(oldId, docCreatedCallback) {

        stopUpdate = true;


        $("#doc_list").jqxListBox('clearFilter');
        var bo = getBlankDocumentObject();
        _currentDocList.unshift(bo);
        var item = $("#doc_list").jqxListBox('insertAt', {
            label: '',
            value: JSON.stringify(bo)
        }, 0);
        $("#doc_list").jqxListBox('selectIndex', 0);
        $("#doc_list").jqxListBox('ensureVisible');


        $("#segment-focus-icon").remove(); // remove focus icon if on

        if (_editor.isClean()) oldId = '-1';



        var docO = getDocumentObject();

        // update file if needed
        if (docO.pth && docO.pth.length > 0) {
            files.saveWmFile(docO.pth, docO.txt);
        }
        saveHistory(docO);

        collection.update({
            _id: oldId
        }, docO, function (err, result) {
            //assert.equal(null, err);
            //assert.equal(1, result);

            console.log("Updating: " + result);
            // insert new

            var blankDoc = getBlankDocumentObject();

            if (_currentProjectId != -1) blankDoc.pid = _currentProjectId;

            collection.insert(blankDoc, function (err, result) {
                //assert.equal(err, null);

                _currentId = result[0]._id.valueOf();
                _currentDoc = blankDoc; // set global _currentDoc

                // create new doc
                _editor.swapDoc(CodeMirror.Doc("", _editor.getOption("mode")));
                // _editor.setValue("");
                _editor.markClean();
                _isEditorSaved = true;
                uiUpdateDocumentName("");

                _editor.focus();

                console.log("Document created with id: " + result[0]._id.valueOf());

                if (docCreatedCallback && typeof docCreatedCallback == "function") {
                    docCreatedCallback();
                }

                // Improve!!!!!!: anatomically update list
                setDocumentList();
            });

        });
    }


    /*
    ########   ########   ##         ########   ########   ########
    ##     ##  ##         ##         ##            ##      ##
    ##     ##  ##         ##         ##            ##      ##
    ##     ##  ######     ##         ######        ##      ######
    ##     ##  ##         ##         ##            ##      ##
    ##     ##  ##         ##         ##            ##      ##
    ########   ########   ########   ########      ##      ########
    */

    function deleteDocument(id) {

        // marks file as deleted; doesn't physically remove it from the collection
        // also implement archive; move document to different collection (speed)
        var elIndex = $("#doc_list").jqxListBox('selectedIndex');
        var item = $("#doc_list").jqxListBox('getItem', elIndex);

        if (_currentDoc.lck == true) {
            wm.showNotice("Can't delete document marked as Read Only.<br><br>Remove Read Only protection first.", {});
            return;
        }


        $("#p5").fadeOut(150).fadeIn(70).fadeOut(70).fadeIn(200);




        $("#doc_list").jqxListBox('unselectItem', item);
        var isLast = ($("#doc_list").jqxListBox('getItems').length == 1);


        if (elIndex != 0) {
            !isLast && $("#doc_list").jqxListBox('selectIndex', elIndex - 1);
            $("#doc_list").jqxListBox('removeAt', elIndex);

        } else {
            $("#doc_list").jqxListBox('removeAt', elIndex);
            !isLast && $("#doc_list").jqxListBox('selectIndex', elIndex);
        }

        isLast && enableEditor(false);


        // drag hack
        $("#doc_list").jqxListBox({
            dragStart: function (item) {
                return false;
            }
        });


        collection.update({
            _id: id
        }, {
            $set: {
                del: true,
                dtm: new Date().getTime()
            }
        }, function (err, numberOfUpdatedDocs) {
            // assert.equal(null, err);
            // assert.equal(1, numberOfUpdatedDocs);
            console.log("Deleting: " + numberOfUpdatedDocs);


            _editor.markClean();
        });
    }


    function unDeleteDocument(id) {

        var elIndex = $("#doc_list").jqxListBox('selectedIndex');
        var item = $("#doc_list").jqxListBox('getItem', elIndex);


        collection.update({
            _id: id
        }, {
                $set: {
                    del: false
                }
            }, function (err, numberOfUpdatedDocs) {

                $("#doc_list").jqxListBox('unselectItem', item);
                var isLast = ($("#doc_list").jqxListBox('getItems').length == 1);


                if (elIndex != 0) {
                    !isLast && $("#doc_list").jqxListBox('selectIndex', elIndex - 1);
                    $("#doc_list").jqxListBox('removeAt', elIndex);

                } else {
                    $("#doc_list").jqxListBox('removeAt', elIndex);
                    !isLast && $("#doc_list").jqxListBox('selectIndex', elIndex);
                }

                isLast && enableEditor(false);

                wm.showNotice('Deleted document restored to its previous location.', {
                    autoClose: true
                });

                _editor.markClean();
            });
    }


    function getNeighborDocIndex(current_index) {
        if ($("#doc_list").jqxListBox('getItem', current_index + 1)) {
            return current_index + 1;
        } else if ($("#doc_list").jqxListBox('getItem', current_index - 1)) {
            return current_index - 1;
        } else {
            return -1;
        }
    }


    /*
    ########   ########   ##     ##   #######   ##     ##  ########
    ##     ##  ##         ###   ###  ##     ##  ##     ##  ##
    ##     ##  ##         #### ####  ##     ##  ##     ##  ##
    ########   ######     ## ### ##  ##     ##  ##     ##  ######
    ##   ##    ##         ##     ##  ##     ##   ##   ##   ##
    ##    ##   ##         ##     ##  ##     ##    ## ##    ##
    ##     ##  ########   ##     ##   #######      ###     ########
    */
    function removeDocument(id) {
        // this will remove permanently
        // just set fields, del or arc
        collection.remove({
            _id: id,
            del: true
        }, {
            //w: 1
        }, function (err, numberOfRemovedDocs) {
            // assert.equal(null, err);
            // assert.equal(1, numberOfRemovedDocs);
            console.log("Removed: " + numberOfRemovedDocs);
        });

    }

    function emptyTrash() {

        // projectsCollection.remove({
        // 	del: true
        // }, {}, function(err, numberOfRemovedDocs) {
        // 	wm.showNotice('<strong>Project removed: </strong>' + numberOfRemovedDocs, {
        // 		autoClose: false
        // 	});
        // });

        collection.remove({
            del: true
        }, {}, function (err, numberOfRemovedDocs) {
            wm.showNotice('<strong>Trash is empty.</strong> (' + numberOfRemovedDocs + ' documents deleted)', {
                autoClose: false
            });

            if ($("#sources_list").jqxTree('getItem', $('#p5')[0]).selected) {
                storage.setDocumentList({
                    del: true
                });
            }

        });

    }

    /*
    ##          #######      ###     ########
    ##         ##     ##    ## ##    ##     ##
    ##         ##     ##   ##   ##   ##     ##
    ##         ##     ##  ##     ##  ##     ##
    ##         ##     ##  #########  ##     ##
    ##         ##     ##  ##     ##  ##     ##
    ########    #######   ##     ##  ########
    */
    var performance0;
    var performance1;
    var performance2;

    function loadDocument(oldId, newId, oldObject) {

        if (newId == '-1') return;

        saveRecentDocuments(newId);


        performance0 = performance.now();

        // check if update is necessary
        // compare new and old hash
        if (_editor.isClean() && isCleanMarks()) {
            // load only
            // var t0 = performance.now();


            collection.findOne({
                    _id: newId
                },
                function (err, document) {

                    // assert.equal(err, null);
                    performance1 = performance.now();
                    // console.log("Doc fetched in: %d ms", (performance1 - performance0));
                    setCM(document);
                });



        } else {

            // update file if needed
            if (oldObject.pth && oldObject.pth.length > 0) {
                files.saveWmFile(oldObject.pth, oldObject.txt);
            }
            saveHistory(oldObject);


            collection.update({
                _id: oldId
            }, oldObject, function (err, result) {


                collection.findOne({
                    _id: newId
                }, function (err, document) {
                    // assert.equal(err, null);
                    performance1 = performance.now();
                    // console.log("Doc fetched in: %d ms", (performance1 - performance0));
                    setCM(document);
                });



            });
        }
    }

    function saveRecentDocuments(docId) {

        if (!localStorage['DOCUMENTS_RECENT_DOCS'])
			localStorage['DOCUMENTS_RECENT_DOCS'] = "[]";
        
        var arr = JSON.parse(localStorage['DOCUMENTS_RECENT_DOCS']);

        if (docId !== arr[0]) {

			if (arr.indexOf(docId) != -1) {
				arr.splice(arr.indexOf(docId), 1);
			}

			arr.unshift(docId);

			if (arr.length > _settings.last_opened_list_max_num) {
				arr.splice(-1, arr.length - _settings.last_opened_list_max_num);
			}
			localStorage['DOCUMENTS_RECENT_DOCS'] = JSON.stringify(arr);
		}


    }

    function removeFromRecentDocuments(docId) {
        var arr = JSON.parse(localStorage['DOCUMENTS_RECENT_DOCS']);
        var elIndex = arr.indexOf(docId);
        if (elIndex !== -1) {
            arr.splice(elIndex, 1);
            localStorage['DOCUMENTS_RECENT_DOCS'] = JSON.stringify(arr);
        }
    }


    /*
    ########   ########   ####  ########    #######   ########
    ##         ##     ##   ##      ##      ##     ##  ##     ##
    ##         ##     ##   ##      ##      ##     ##  ##     ##
    ######     ##     ##   ##      ##      ##     ##  ########
    ##         ##     ##   ##      ##      ##     ##  ##   ##
    ##         ##     ##   ##      ##      ##     ##  ##    ##
    ########   ########   ####     ##       #######   ##     ##
    */


    function clearCM() {}


    function setCM(docO) {

        var saveAfter = false;

        // first check if bound to file and if file is newer
        // console.log(docO.pth)

        if (docO.pth && docO.pth.length > 0) { // is file

            var fl = files.getWmFile(docO.pth); // get file


            if (fl == null) { // file not present

                // file not found, ask to recreate
                // decouple

                wm.showNotice('File:<br><strong>' + docO.pth + '</strong><br><br>bound to this document has been deleted, moved or renamed!<br><br>', {
                    autoClose: false
                });
                saveAfter = true;
                docO.pth = "";

            } else { // file is there, check hashes

                if (docO.hsh !== fl.hash) { // files are different, ask user which version to keep

                    var modal = new tingle.modal({
                        footer: true,
                        stickyFooter: false,
                        closeMethods: [],
                        closeLabel: "",
                        beforeClose: function () {
                            return true; // close the modal
                        }
                    });

                    // check which is newer
                    var newerOrOlder = (moment(docO.dtm).isBefore(fl.stats.mtime)) ? "(newer)" : "(older)";

                    var actionLabel = "Select which version do you want to keep.<br><br><strong>Be careful, data loss my occur! If you are not sure select <em>Keep local & unbind file</em>. Both versions of text will be preserved.</strong>";

                    var readOnlyLabel = docO.lck ? "<br><br><strong>Document is marked as Read only and it can't be updated from the file!</strong>" : "";

                    // show modal dialog
                    var modalText = '<div class="tingle-content-wrapper">File:<strong>' + docO.pth + '</strong><br>bound to this document is different ' + newerOrOlder + ' than local database version.<br><br>It seems it was modified outside Writemonkey.<br><br>File last saved: <strong>' + moment(fl.stats.mtime).format('MMMM Do YYYY, h:mm:ss a') + '</strong> (<strong>' + stats.getWords(fl.text).count + '</strong> words)<br>Local last saved: <strong>' + moment(docO.dtm).format('MMMM Do YYYY, h:mm:ss a') + '</strong> (<strong>' + stats.getWords(docO.txt).count + '</strong> words)<br></span><br><br>' + actionLabel + readOnlyLabel + '</div>';


                    modal.setContent(modalText);

                    // Unbind
                    modal.addFooterBtn('Keep local & unbind file', 'tingle-btn', function () {
                        _currentDoc.pth = "";
                        _editor.execCommand("save"); // rewrite file with doc contents
                        modal.close();
                    });



                    // Keep database
                    modal.addFooterBtn('Keep local & overwrite file', 'tingle-btn', function () {
                        _editor.execCommand("save"); // rewrite file with doc contents
                        modal.close();
                    });


                    if (!docO.lck) {
                        // Keep file
                        modal.addFooterBtn('Update from file', 'tingle-btn', function () {
                            // reset
                            _currentDoc.txt = fl.text;
                            _currentDoc.bkm = [];
                            _currentDoc.mrk = [];
                            _currentDoc.car = {
                                line: 0,
                                ch: 0
                            };
                            _currentDoc.fld = [];
                            _currentDoc.und = {};
                            _currentDoc.hsh = fl.hash;
                            _currentDoc.lck = false;

                            _editor.setValue(fl.text);

                            modal.close();
                        });
                    }


                    // wm theme styling for modal dialogs - tingle
                    $(".tingle-btn").css({
                        "background": localStorage["VIEWPORT_MARK_COLOR"],
                        "color": localStorage["VIEWPORT_BACKGROUND_COLOR"]
                    });

                    modal.open();
                }
            }
        }


        // finish loading document
        _editor.operation(function () {

            wm.animateProgressBar(); // animate progress bar
            $("#segment-focus-icon").remove(); // remove focus icon if on


            _editor.swapDoc(CodeMirror.Doc(docO.txt, _editor.getOption("mode")));
            _editor.setBookmarks(docO.bkm);
            _editor.setCollapsed(docO.mrk);
            _lastChangePosition = docO.car;
            _editor.setFolds(docO.fld);
            _editor.execCommand("showScrollBar");

            if (!$.isEmptyObject(docO.und))
                _editor.setHistory(docO.und);

            _editor.refresh();
            _editor.markClean();

            stopUpdate = false;

            _currentId = docO._id.valueOf();
            _currentDoc = docO; // set helper global
            _isEditorSaved = true;
            uiUpdateDocumentName(docO.nam);

            // jump to last edit if enabled
            var setCursorTo = (_settings.jump_to_last_edit_on_document_open === true) ? docO.car : {line: 0, ch: 0};
            _editor.setCursor(setCursorTo, {
                scroll: true
            });
            _editor.execCommand("scrollCursorToTypewriterPos");
            jumps.showJumpSpot(setCursorTo.line);


        });



        // remove search matches and show hits if search is enabled
        _editor.execCommand("clearSearch");
        _editor.removeMatches();
        if ($('#doc_search_input').val() != "") {
            _editor.showMatches($('#doc_search_input').val().replace(/_/g, ' ').replace(/#|%/g, '').trim());
        }

        // for read only
        if (docO.lck == true) {
            _editor.enableReadOnly();
        }

        // save if after update from file
        if (saveAfter) {
            _editor.execCommand("save");
        }

    }



    /*
    ##     ##  ########   ########      ###     ########   ########
    ##     ##  ##     ##  ##     ##    ## ##       ##      ##
    ##     ##  ##     ##  ##     ##   ##   ##      ##      ##
    ##     ##  ########   ##     ##  ##     ##     ##      ######
    ##     ##  ##         ##     ##  #########     ##      ##
    ##     ##  ##         ##     ##  ##     ##     ##      ##
     #######   ##         ########   ##     ##     ##      ########
    */

    function updateFields(id, fieldsObject) {
        collection.update({
            _id: id
        }, {
            $set: fieldsObject
        }, function (err, numberOfUpdatedDocs) {
            // assert.equal(null, err);
            // assert.equal(1, numberOfUpdatedDocs);
        });
    }



    // Update document based on id
    function updateDocument(id) {

        var docO = getDocumentObject();
        docO._id = id;
        var selectedItem = $("#doc_list").jqxListBox('getSelectedItem');
        _currentDoc = docO;

        // update file if needed
        if (_currentDoc.pth && _currentDoc.pth.length > 0) {
            files.saveWmFile(_currentDoc.pth, _currentDoc.txt);
        }
        // save history
        saveHistory(docO);


        // Update list item without refreshing
        docO.met = (docO.met) ? docO.met : {};
        docO.met.tags = (docO.met.tags) ? docO.met.tags : [];

        $("#doc_list").jqxListBox('updateAt', {
            label: docO.nam,
            value: JSON.stringify(docO)
        }, $("#doc_list").jqxListBox('getSelectedIndex'));

        $("#doc_list").jqxListBox('ensureVisible', selectedItem);
        _isEditorSaved = true;
        uiUpdateDocumentName(docO.nam);


        // Now do the real database update
        collection.update({
            _id: id
        }, docO, function (err, numberOfUpdatedDocs) {

            console.log("Updating: " + id);
            _editor.markClean();
            $("#doc_list").jqxListBox('invalidate');
            saveRecentDocuments(id);
        });

    }

    // Check if marked regions changed since last save --> compliments _editor.isClean() function
    function isCleanMarks() {
        var currentMarks = _editor.getMarks();

        function getIsCleanMarksIdString(bkm, fld, mrk) {
            return JSON.stringify(bkm) + JSON.stringify(fld) + JSON.stringify(mrk);
        }

        if (_currentDoc == null) {
            return true;
        } else if (getIsCleanMarksIdString(_currentDoc.bkm, _currentDoc.fld, _currentDoc.mrk) == getIsCleanMarksIdString(currentMarks.bookmarks, currentMarks.folds, currentMarks.collapsed)) {
            return true;
        }
        return false;
    }




    /*
    ##     ##  ####   ######    ########    #######   ########   ##    ##
    ##     ##   ##   ##    ##      ##      ##     ##  ##     ##   ##  ##
    ##     ##   ##   ##            ##      ##     ##  ##     ##    ####
    #########   ##    ######       ##      ##     ##  ########      ##
    ##     ##   ##         ##      ##      ##     ##  ##   ##       ##
    ##     ##   ##   ##    ##      ##      ##     ##  ##    ##      ##
    ##     ##  ####   ######       ##       #######   ##     ##     ##
    */



    function saveHistory(obj) {

        if (!_settings.keep_document_history)
            return;

        var hobj = {
            doc_id: _currentId,
            event_date: new Date().getTime(),
            name: obj.nam,
            text: obj.txt,
            words: obj.wrd,
            cursor: obj.car
        };

        historyCollection.insert(hobj, function (err, result) {
            console.log("History saved");
        });
    }


    function getDocumentHistoryList(id, callback) {
        historyCollection
            .find({
                doc_id: id
            }) // { 'nam':true,'dtm':true, 'wrd':true,  }
            .sort({
                event_date: -1
            })
            .limit(1000)
            .toArray(function (err, historyArray) {
                callback(historyArray);
            });
    }

    function showDocumentHistoryPalette() {

        getDocumentHistoryList(_currentId, function (historyArray) {
            var paletteA = [];

            for (var i = 0; i < historyArray.length; i++) {

                var timestamp = "<strong>" + moment(historyArray[i].event_date).format('MMM DD HH:mm') + "</strong>";
                var label = '<span style="font-size: .9em;">' + historyArray[i].text.substr(historyArray[i].text.length - 45).trim() + '</span>';

                paletteA.push({
                    id: '',
                    string: timestamp + " ..." + label,
                    shortcut: stats.add000separator(historyArray[i].words) + " words",
                    action: 'storage.restoreDocumentFromHistory(' + historyArray[i]._id.toString() + ')',
                    type: "eval",
                    sublist: [],
                    persist: false
                });
            }
            commandPalette.hideCommandPalette();
            commandPalette.showCommandPalette(paletteA);
            $('.searchFilter').val("");
        });
    }


    function restoreDocumentFromHistory(id) {

        historyCollection.findOne({
            _id: id
        }, function (err, document) {
            _editor.setValue(document.text);
            _editor.setCursor(document.cursor);
            _editor.refresh();

            wm.showNotice("Restored from history! Use Undo to revert.", {
                autoClose: false
            });
        });
    }




    /*
       ###     ##     ##  ########    #######    ######       ###     ##     ##  ########
      ## ##    ##     ##     ##      ##     ##  ##    ##     ## ##    ##     ##  ##
     ##   ##   ##     ##     ##      ##     ##  ##          ##   ##   ##     ##  ##
    ##     ##  ##     ##     ##      ##     ##   ######    ##     ##  ##     ##  ######
    #########  ##     ##     ##      ##     ##        ##   #########   ##   ##   ##
    ##     ##  ##     ##     ##      ##     ##  ##    ##   ##     ##    ## ##    ##
    ##     ##   #######      ##       #######    ######    ##     ##     ###     ########
    */

    var autoSaveInterval = null;
    var autoSaveDelayTimeout = null;

    var isEditorInactive = true;
    var secondsFromLastSave = 0;
    var intervalInSec = parseInt(_settings.auto_save_interval, 10) * 60;
    var isAutoSaveOn = intervalInSec >= 60;



    function setAutoSave() {
        if (isAutoSaveOn) { // auto save is enabled
            autoSaveInterval = setInterval(tickAutoSave, 1000);
        }
    }


    function tickAutoSave() {

        if (!_editor.isClean()) secondsFromLastSave++; // step seconds counter
        // check if auto save should fire
        if (secondsFromLastSave >= intervalInSec &&
            isEditorInactive &&
            _currentId != -1 &&
            !_editor.isClean()) {
            console.log('auto save');
            _editor.execCommand("save");
        }

    }

    function resetAutoSave() {
        secondsFromLastSave = 0;
    }


    document.body.addEventListener("saveE", function () {
        resetAutoSave();
    }, false);


    _editor.on("swapDoc", function (cm, oldDoc) {
        resetAutoSave();
    });


    _editor.on("change", function () {
        if (!isAutoSaveOn) return;

        isEditorInactive = false;
        clearTimeout(autoSaveDelayTimeout);
        autoSaveDelayTimeout = setTimeout(function () {
            isEditorInactive = true;
        }, 3000);
    });



    /*
     ######    ########   ########   ##         ####   ######    ########
    ##    ##   ##            ##      ##          ##   ##    ##      ##
    ##         ##            ##      ##          ##   ##            ##
     ######    ######        ##      ##          ##    ######       ##
          ##   ##            ##      ##          ##         ##      ##
    ##    ##   ##            ##      ##          ##   ##    ##      ##
     ######    ########      ##      ########   ####   ######       ##
    */
    function setDocumentList(query, sort) {

        var s, q;



        if (!query) {
            if (_currentProjectId != -1) {
                q = {
                    del: false,
                    pid: _currentProjectId
                };
            } else {
                q = {
                    del: false
                };
            }
        } else {
            q = query;
        }



        if (!sort) {
            if (_currentProjectId != -1) {
                s = {
                    ord: -1,
                    dtm: -1
                };
            } else {
                s = {
                    dtm: -1
                };
            }
        } else {
            s = sort;
        }


        collection
            .find(q, {
                txt: 0,
                und: 0
            })
            .sort(s)
            .limit(10000)
            .toArray(function (err, docArr) {
                wisdom.stopSpinner();
                $("#doc_list").jqxListBox("clearFilter");


                enableEditor(true);

                if (docArr.length <= 0) { // no docs on list

                    enableEditor(false);

                    // ensure at least one empty doc when appropriate
                    if (_currentProjectId != -1 || $("#sources_list").jqxTree('getItem', $('#p1')[0]).selected) {

                        if ($("#doc_search_input").val() == "") {
                            newDocument(-1);
                            return;
                        }
                    }
                }

                createListBox(docArr);
            });
    }


    function enableEditor(state) {
        if (state == true) {
            $("#main_pane, #module_pane").css("opacity", "1");
            _editor.setOption("readOnly", false);
        } else {
            $("#main_pane, #module_pane").css("opacity", "0");
            _editor.setOption("readOnly", "nocursor");
        }
    }



    function doFullTextSearch(input) {

        if (input == "") { // select all and exit
            $("#sources_list").jqxTree('selectItem', $('#p1')[0]);
            return;
        }

        var andA = [];
        var orA = [];

        // hack to refere to and or or
        var or_and = {
            ora: orA,
            anda: andA
        }
        var referer = "anda"; // default is AND

        input = input.trim();

        // set folder only
        // if (input.startsWith("%") && _currentProjectId != -1) {
        //     input = input.replace(/^%\s*/, "");
        //     andA.push({
        //         pid: _currentProjectId
        //     });
        // }


        // set or condition
        if (input.startsWith("or:")) {
            input = input.substring(3).trim();
            referer = "ora" // set OR
        }

        // tokenize input
        var inputTokens = input.split(" ");


        for (var i = 0; i < inputTokens.length; i++) {

            var token = inputTokens[i].trim();

            if (token.startsWith("$")) { // search on name field only
                or_and[referer].push({
                    nam: {
                        $regex: token.substring(1).replace(/_/g, ' '),
                        $options: 'gim'
                    }
                });

            } else if (token.startsWith("#")) { // search tags only
                or_and[referer].push({
                    'met.tags': {
                        $regex: '^' + token.substring(1) + '$',
                        $options: 'i'
                    }
                });

            } else { // serach on txt by default
                or_and[referer].push({
                    txt: {
                        $regex: token.replace(/_/g, ' '),
                        $options: 'gim'
                    }
                });
            }
        } // end for loop


        // add del for all queries / never search deleted items
        andA.push({
            del: false
        });


        $("#sources_list").jqxTree('selectItem', $('#p8')[0]);
        saveSearch($('#doc_search_input').val());


        setDocumentList({
            $and: andA,
            $or: orA
        });

    }

    function saveSearch(input) {

        if (!localStorage['SEARCH_LAST_ITEMS'])
            localStorage['SEARCH_LAST_ITEMS'] = "[]";

        var arr = JSON.parse(localStorage['SEARCH_LAST_ITEMS']);
        arr.unshift(input);
        if (arr.length >= 16) {
            arr.splice(-1, 1);
        }
        arr = arr.filter(function (item, pos) {
            return arr.indexOf(item) == pos;
        })

        localStorage['SEARCH_LAST_ITEMS'] = JSON.stringify(arr);

    }


    function getLastSearchesArray() {
        return JSON.parse(localStorage['SEARCH_LAST_ITEMS']);
    }





    /*
     #######   ########         ##
    ##     ##  ##     ##        ##
    ##     ##  ##     ##        ##
    ##     ##  ########         ##
    ##     ##  ##     ##  ##    ##
    ##     ##  ##     ##  ##    ##
     #######   ########    ######
    */
    function getDocumentObject() {

        var meta = getMetaFields();
        var name = getDocumentName();
        var marks = _editor.getMarks();
        var value = _editor.getValue();
        var hash = files.getHashFromString(value);

        return {
            dtc: (_currentDoc && _currentDoc.dtc) ? _currentDoc.dtc : 0,
            dtm: (new Date().getTime()), // date modified
            pth: (_currentDoc && _currentDoc.pth) ? _currentDoc.pth : "", // path if is file
            met: meta, // document meta data
            ses: (_currentDoc && _currentDoc.ses && _currentDoc.ses.startTime) ? _currentDoc.ses : {}, // time spent on file {start: 0, stop: 0}
            scr: false, // is scratch
            pid: (_currentDoc && _currentDoc.pid) ? _currentDoc.pid : "", // project id
            ord: (_currentDoc && _currentDoc.ord) ? _currentDoc.ord : 0,
            nam: name, // doc name - auto generated
            //tgs: [], // document tags
            tsk: getTasks(), // to-do items in text
            bkm: marks.bookmarks, // list of bookmarks
            mrk: marks.collapsed, // list of collapsed regions
            txt: value, // document main text
            sum: "", // summary - common words with frequencies
            wrd: stats.getWords(value).count, // word count
            exp: "", // export template
            car: _editor.getCursor(), // caret position
            sel: [], // selections
            fld: marks.folds, // folds
            und: _settings.keep_document_undo_data ? _editor.getHistory() : {}, // undo history
            fav: (_currentDoc && _currentDoc.fav) ? _currentDoc.fav : false, // is favourite
            del: (_currentDoc && _currentDoc.del) ? _currentDoc.del : false, // is deleted
            arc: false, // is archived
            rep: (_currentDoc && _currentDoc.rep) ? _currentDoc.rep : false, // is repository text; scrap doc
            lck: (_currentDoc && _currentDoc.lck) ? _currentDoc.lck : false, // is locked, read only
            opt: {}, // addittional options and settings
            hsh: hash //getHash(_editor.getValue()) // unique text hash - main + repo
        };
    }

    // refactor
    function getBlankDocumentObject() {
        var timestamp = new Date().getTime();

        return {
            dtc: timestamp, // date created - only set first time, don't update!!
            dtm: timestamp, // date modified
            pth: "", // path if is file
            met: {}, // document meta data
            ses: {}, // time spent on file {start: 0, stop: 0}
            scr: true, // is scratch
            pid: "", // project id
            ord: timestamp, // order in collection / project
            nam: "SCRATCH", // doc name - auto generated
            //tgs: [], // document tags
            tsk: {}, // to-do items in text
            bkm: [], // list of bookmarks
            mrk: [], // list of collapsed regions
            txt: "", // document main text
            sum: "", // summary - common words with frequencies
            wrd: 0, // word count
            exp: "", // export template
            car: {
                line: 0,
                ch: 0
            }, // caret position
            sel: [], // selections
            fld: [], // folds
            und: {}, // undo history
            fav: false, // is favourite
            del: false, // is deleted
            arc: false, // is archived
            rep: false, // is repository text; scrap doc
            lck: false, // is locked
            opt: {}, // addittional options and settings
            hsh: 0 // unique text hash - main + repo,
        };
    }


    /*
    ##     ##  ########   ##         ########        ########   ####  ##         ########
    ##     ##  ##         ##         ##     ##       ##          ##   ##         ##
    ##     ##  ##         ##         ##     ##       ##          ##   ##         ##
    #########  ######     ##         ########        ######      ##   ##         ######
    ##     ##  ##         ##         ##              ##          ##   ##         ##
    ##     ##  ##         ##         ##              ##          ##   ##         ##
    ##     ##  ########   ########   ##              ##         ####  ########   ########
    */
    // function insertHelpIfBlank() {
    // 	var helpText = fs.readFileSync("README.md", {
    // 		encoding: "utf-8"
    // 	});

    // 	var obj = getBlankDocumentObject();
    // 	obj.txt = helpText;
    // 	obj.nam = "WM3 Help file"

    // 	collection.insert(obj, function(err, result) {});

    // }


    /*
    ########    #######   ##     ##
    ##     ##  ##     ##  ###   ###
    ##     ##  ##     ##  #### ####
    ##     ##  ##     ##  ## ### ##
    ##     ##  ##     ##  ##     ##
    ##     ##  ##     ##  ##     ##
    ########    #######   ##     ##
    */



    function createListBox(arr) {

        // arr = arr.map(function(item) {
        // 	item.met = (item.met) ? item.met : {}; // if there is no met item in old entry
        // 	item.met.tags = (item.met.tags) ? item.met.tags : [];
        // 	var tagsString = tagsStringFromArray(item.met.tags);
        // 	var fileName = path.basename(item.pth);
        // 	item.filterString = item.nam + "%*%" + tagsString + " " + fileName;
        // 	return item;
        // });


        var valueArr = arr.map(function (item) {

            return {
                id: item._id,
                label: item.nam,
                value: JSON.stringify(item)
            };
        });


        _currentDocList = arr;


        var t0 = performance.now();


        $('#doc_list').jqxListBox("clear");



        var source = {
            datatype: "array",
            datafields: [{
                name: 'id',
                type: 'int'
            }, {
                name: 'label',
                type: 'string'
            }, {
                name: 'value',
                type: 'string'
            }],
            id: 'id',
            localdata: valueArr
        };

        var settings = {
            async: false,
            autoBind: true
        };
        var dataAdapter = new $.jqx.dataAdapter(source, settings);


        $("#doc_list").jqxListBox({
            theme: "metro",
            selectedIndex: 0,
            autoItemsHeight: true,
            equalItemsWidth: true,
            enableSelection: true,
            enableHover: true,
            filterable: false,
            filterDelay: 100,
            filterHeight: 35,
            filterPlaceHolder: "<filter>",
            searchMode: "containsignorecase",
            checkboxes: false,
            incrementalSearch: false,
            multiple: false,
            multipleextended: false,
            scrollBarSize: 4,
            itemHeight: -1,
            autoHeight: false, // must be false for scrollbars to work
            allowDrop: (_currentProjectId != -1) ? true : false,
            allowDrag: true, //(_currentProjectId != -1) ? true : false,
            source: dataAdapter,
            displayMember: "label",
            valueMember: "value",

            renderer: function (index, label, value) {

                var datarecord = JSON.parse(value);

                if (_currentId == datarecord._id) { // get updated object
                    var newO = getDocumentObject();
                    newO._id = datarecord._id;
                    datarecord = newO;
                    _currentDocList[index] = newO;
                }


                if (datarecord) {



                    var tx = new moment();
                    var ty = new moment(datarecord.dtm);
                    var duration = moment.duration(tx.diff(ty, 'days'));
                    var ago = (duration > 28) ? ty.format("ll") : ty.fromNow();


                    var prc = (datarecord.wrd / datarecord.met.goal * 100).toFixed(0);
                    var goal = (datarecord.met && datarecord.met.goal) ? " (" + prc + "%)" : "";


                    var count = "words: " + stats.add000separator(datarecord.wrd) + goal;

                    var lbl = datarecord.nam == "" ? "SCRATCH" : datarecord.nam;

                    var nmb = _settings.number_before_document_name ? `<tag style="font-weight: normal; border-width: 0px;">${index + 1}. </tag>` : "";


                    var tags = (datarecord.met && datarecord.met.tags) ? datarecord.met.tags : [];

                    tags = tags.map(function (s) {

                        var text = s;
                        var opa = .6;
                        var cl = "";
                        var size = ".67rem";
                        var mar = ".15rem";
                        var mar2 = "0px";
                        var gry = "0";

                        if (isValidColor(s)) {
                            text = "&#x25CF;"; // &#xf004; &#xf005; &#xf02e; &#xf0c8; &#xf04d &#xf04d;
                            opa = 1;
                            cl = "awf";
                            size = "1rem";
                            mar = ".15rem";
                            mar2 = "0px";
                            gry = ".3";
                        }



                        return `<tag class="${cl}" style="width: auto; height: auto; margin-right: ${mar}; margin-left: ${mar2};font-size: ${size}; padding: 0px 1px 0px 0px; opacity: ${opa}; font-weight: normal; color: ${s}; border: 0px solid ${s};">` + text + '</tag>';
                    });



                    tags = tags.join("");
                    //tags = tags.length > 0 ? tags : "";

                    // get tasks count
                    var taskBadge = "";
                    if (datarecord.tsk && datarecord.tsk.ndn) {
                        var todos = datarecord.tsk.ndn.length;
                        if (todos > 0) {
                            taskBadge = `<tag class="theme-background-color-mark theme-text-color-inverse" style="min-width: .9rem; height: .9rem; line-height: .9rem; border: 1px solid; text-align: center; float: right; margin-top: 0px; padding: 0px 0px 0px 0px; font-size: .55rem; opacity: .7; border-radius: .9rem;">${todos}</tag>`
                        }
                    }



                    // get done graph
                    // var progressBadge = "";
                    // var prcnorm = prc > 100 ? 100 : prc;
                    // if (datarecord.met && datarecord.met.goal) {
                    //     progressBadge = `<tag class="theme-border-color-inverse" style="background: -webkit-linear-gradient(left, rgba(127,127,127,0.3) ${prcnorm}%, transparent ${prcnorm}%); border: 1px solid; min-width: 35px; height: 5px; line-height: 5px; text-align: center; float: right; margin-top: 3px; padding: 0px 0px 0px 0px; font-size: .6em; opacity: .7; background-color: transparent; border-radius: 5px;"></tag>`
                    // }

                    // set file bind badge
                    var fileBadge = "";
                    if (datarecord.pth && datarecord.pth.length > 0) {
                        fileBadge = '<tag id="file-bind-' + datarecord._id + '" style="opacity: .8; font-family: Font Awesome">&#xf0c1;&nbsp;</tag>';
                    }

                    // set fav bind badge
                    var favBadge = "";
                    if (datarecord.fav && datarecord.fav == true) {
                        favBadge = '<tag class="document-list-favourite-mark" id="is-fav-' + datarecord._id + '" style="opacity: .8; font-family: Font Awesome">&#xf005;&nbsp;</tag>';
                    }

                    // set locked badge
                    var lckBadge = "";
                    if (datarecord.lck && datarecord.lck == true) {
                        lckBadge = '<tag class="document-list-readonly-mark" id="is-readonly-' + datarecord._id + '" style="opacity: .8; font-size: 1.1em; font-family: Font Awesome">&#xf023;&nbsp;</tag>';
                    }

                    // set is scrap
                    var scrap = "";
                    if (datarecord.rep == true) {
                        scrap = 'opacity: .4;'
                    }

                    if (localStorage['VIEWPORT_DOCUMENT_PANE_COMPACT'] == 1) {
                        var tile = `<div id="document-tile-${datarecord._id}" class="document-tile" style="padding: 0px; overflow: hidden; height: 1rem; background: transparent;"><div style="overflow: hidden; text-overflow: ellipsis; background: transparent; white-space: nowrap; font-size: .83rem; margin-top:0px; ${scrap}">${lckBadge}${favBadge}${fileBadge}${nmb}${lbl}</div></div>`;

                    } else {
                        lbl = `<strong>${lbl}</strong>`;
                        var tile = `<div id="document-tile-${datarecord._id}" class="document-tile" style="padding: 0px; overflow: hidden; height: 4.9rem; ${scrap}">` +
                            `<tag style="font-size: .7rem; margin-bottom:.6rem;">${ago}</tag>&nbsp;${taskBadge}` +
                            `<div style="overflow: hidden; text-overflow: ellipsis; white-space: nowrap; font-size: .83rem; height:1.2rem;  margin-top:.3rem; font-weight: normal;">${lckBadge}${favBadge}${fileBadge}${nmb}${lbl}</div>` +
                            `<div style="font-size: .7rem; margin-top: .2rem;">${count}</div>` +
                            `<div style="font-size: .7rem; margin-top: .45rem; height: .7rem; float: right;">${tags}</div>` + `</div>`;
                            // `<div style="font-size: .7rem; margin-top: .45rem; float: left;">${progressBadge}</div>`
                    }

                    return tile;
                }
            },
            width: "100%",
            height: "100%"
        });

        $("#doc_list").jqxListBox('selectIndex', 0);
        stopUpdate = false;

        uiUpdateDocumentName(getDocumentName());


    } // END create listbox



    function uiUpdateDocumentName(name) {
        name = (name == "") ? "SCRATCH" : name;
        name = (name.length > 30) ? name.substring(0, 30) + "..." : name;

        var fileLbl = "";
        var readOnlyLbl = "";
        var fileNam = "";
        var projectLabel = "";
        var notSavedIndicator = "";

        if (_settings.show_not_saved_indicator && !_isEditorSaved) {
            notSavedIndicator = '<span style="font-size: 1em; margin-left: .3em;">&#9679;</span>';
        }


        if (_currentDoc && _currentDoc.pth && _currentDoc.pth.length > 0) {
            fileLbl = '<span style="font-family: Font Awesome; font-size:.8em; margin-left: 3px;">&#xf0c1;</span>';
            fileNam = " [" + path.basename(_currentDoc.pth) + "]";
        }

        if (_currentDoc && _currentDoc.lck == true) {
            readOnlyLbl = '<span style="font-family: Font Awesome; font-size:.9em; margin-left: 3px;">&#xf023;</span>';
        }

        if (_currentDoc && _currentProjectsList[_currentDoc.pid.toString()]) {
            projectLabel = _currentProjectsList[_currentDoc.pid.toString()].nam + ':&nbsp;&nbsp;';
        }

        $("title").html(projectLabel + name + " - Writemonkey");

        $("#info_bar_middle").html('<span id="info_bar_project">' + projectLabel + '</span><span id="info_bar_name">' + name + fileLbl + readOnlyLbl + '</span>' + notSavedIndicator);

    }


    // update custom order of document list
    function updateCustomOrder(newIndex) {
        // console.log(newIndex);

        var currentValue = JSON.parse($("#doc_list").jqxListBox('getItem', newIndex).value);
        var idToUpdate = currentValue._id;


        // get order from previous
        var prevOrd = new Date().getTime(); // if 0 (to top)

        if (newIndex != 0) {
            var prevValue = JSON.parse($("#doc_list").jqxListBox('getItem', newIndex - 1).value);
            prevOrd = prevValue.ord == 0 ? prevValue.dtm : prevValue.ord;
        }

        // get order from next
        var numOnList = $("#doc_list").jqxListBox('getItems').length;
        var nextOrd = 0; // if last (to bottom)

        if (newIndex < numOnList - 1) {
            var nextValue = JSON.parse($("#doc_list").jqxListBox('getItem', newIndex + 1).value);
            nextOrd = nextValue.ord == 0 ? nextValue.dtm : nextValue.ord;
        }


        var newOrd = prevOrd - ((prevOrd - nextOrd) / 2);
        _currentDoc.ord = newOrd; // update document object


        collection.update({
            "_id": idToUpdate
        }, {
            $set: {
                "ord": newOrd
            }
        }, {
            "multi": false
        }, function (err, numberOfUpdatedDocs) {
            // after updating in database also update jqlistbox item value with newOrd
            var newValue = JSON.parse($("#doc_list").jqxListBox('getItem', newIndex).value);
            newValue.ord = newOrd;
            $("#doc_list").jqxListBox('updateAt', {
                label: newValue.nam,
                value: JSON.stringify(newValue)
            }, $("#doc_list").jqxListBox('getSelectedIndex'));

        });
    }



    /*
    ########   ##     ##  ########   ##    ##   ########    ######
    ##         ##     ##  ##         ###   ##      ##      ##    ##
    ##         ##     ##  ##         ####  ##      ##      ##
    ######     ##     ##  ######     ## ## ##      ##       ######
    ##          ##   ##   ##         ##  ####      ##            ##
    ##           ## ##    ##         ##   ###      ##      ##    ##
    ########      ###     ########   ##    ##      ##       ######
    */

    var idToSelect = -1;
    var lateDocLoader = null;
    var loadPreventInterval = null;
    var loaderTimer = 10;

    function updateDoc() {
        if (idToSelect != -1 && _currentId != idToSelect && !stopUpdate) {
            stopUpdate = true;
            loadDocument(_currentId, idToSelect, getDocumentObject());
            if (lateDocLoader) clearInterval(lateDocLoader);
            // console.log("late load");
        }
    }


    $("#doc_list").on('select', function (event) {
        // drag hack
        $("#doc_list").jqxListBox({
            dragStart: function (item) {
                return true;
            }
        });


        if (loadPreventInterval) clearInterval(loadPreventInterval);
        loadPreventInterval = setInterval(function () {
            loaderTimer++;
        }, 100);

        if (lateDocLoader) clearInterval(lateDocLoader);
        lateDocLoader = setInterval(updateDoc, 300);

        if (!event.args.item) return; // wrong event!

        var item = event.args.item;
        var ob = JSON.parse(item.value);


        idToSelect = (ob._id) ? ob._id : -1;


        if (_currentId != idToSelect && !stopUpdate && loaderTimer > 5) {
            // console.log("primary load");
            loadDocument(_currentId, idToSelect, getDocumentObject());
            stopUpdate = true;
            loaderTimer = 0;
            var opa = $("#info_bar_top").css("opacity");
            $("#info_bar_top_right").fadeTo(1, .001).delay(1500).fadeTo(500, opa);
        }
    });


    $("#doc_list").on('dragStart', function (event) {
        stopUpdate = true;
        loaderTimer = 0;
    });

    $("#doc_list").on('dragEnd', function (event) {
        stopUpdate = false;

        var ev = event.args;
        var oev = ev.originalEvent;

        updateCustomOrder($("#doc_list").jqxListBox("getSelectedIndex"));


    });


    $("#new_doc").on("click", function () {
        newDocument(_currentId); // pass current doc if it needs to be saved
    });

    $("#delete_doc").on("click", function () {
        deleteDocument(_currentId);
    });

    $("#doc_list").on("contextmenu", function (e) {
        e.preventDefault();
        e.stopPropagation();
        // e.stopImmediatePropagation();

        var id = $(event.target).closest(".document-tile").attr("id");
        if (id) {
            menus.getDocumentTileMenu(id).popup(e.originalEvent.clientX, e.originalEvent.clientY + 10);

            // drag hack
            $("#doc_list").jqxListBox({
                dragStart: function (item) {
                    return false;
                }
            });
        }
    });

    $("#doc_list").on("dblclick", function (e) {
        e.preventDefault();
        e.stopPropagation();
        // e.stopImmediatePropagation();
        _editor.execCommand("toggleFoldAll");
    });



    $("#sources_list").on("contextmenu", function (e) {
        // e.preventDefault();
        // e.stopPropagation();
        // e.stopImmediatePropagation();

        var id = $(event.target).closest("li").attr("id");
        if (id) {
            $("#sources_list").jqxTree('selectItem', $('#' + id.toString())[0]);

            if (id == "pb0" || id == "p1" || id == "p2" || id == "p3" || id == "pb7")
                return;

            //if ($('#sources_list').jqxTree('getSelectedItem').id != "10")
            menus.getProjectTileMenu(id).popup(e.originalEvent.clientX, e.originalEvent.clientY + 10);
        }
    });

    $('#sources_list').on('expand', function (event) {

        var item = $('#sources_list').jqxTree('getItem', event.args.element);

        if (item.id.startsWith("pb")) return;

        var ob = localStorage['PROJECTS_TREE_EXPANDED_DATA'] ? JSON.parse(localStorage['PROJECTS_TREE_EXPANDED_DATA']) : {};
        ob[item.id] = true;
        localStorage['PROJECTS_TREE_EXPANDED_DATA'] = JSON.stringify(ob);

    });

    $('#sources_list').on('collapse', function (event) {
        var item = $('#sources_list').jqxTree('getItem', event.args.element);

        if (item.id.startsWith("pb")) return;

        var ob = localStorage['PROJECTS_TREE_EXPANDED_DATA'] ? JSON.parse(localStorage['PROJECTS_TREE_EXPANDED_DATA']) : {};
        ob[item.id] = false;
        localStorage['PROJECTS_TREE_EXPANDED_DATA'] = JSON.stringify(ob);

    });



    // SEARCH

    $("#doc_search_input").on('keyup', function (e) {
        if (e.keyCode == 13) {
            doFullTextSearch($('#doc_search_input').val());
        }
    });

    // $('#doc_search_input').on('dblclick', function() {
    // 	$('#doc_search_input').val("");
    // });


    $('#doc_search_input').on('contextmenu', function (e) {
        menus.getLastSearchMenu(i).popup(e.originalEvent.clientX, e.originalEvent.clientY + 10);
    });

    // var docSearchTimer = null;
    // $('#doc_search_input').on('input', function() {
    // 	if (docSearchTimer) window.clearTimeout(docSearchTimer);
    // 	docSearchTimer = setTimeout(function() {
    // 		setDocumentList();
    // 	}, 500);
    // });

    // $('#doc_search_input').on('dblclick', function() {
    // 	$('#doc_search_input').val("");
    // 	setDocumentList();
    // });

    // $('#doc_header').on('dblclick', function() {
    // 	$("#doc_list").jqxListBox("clearFilter");
    // 	$("#jumps_list").jqxListBox("clearFilter");
    // });





    /*
    ##     ##  ########   ####  ##
    ##     ##     ##       ##   ##
    ##     ##     ##       ##   ##
    ##     ##     ##       ##   ##
    ##     ##     ##       ##   ##
    ##     ##     ##       ##   ##
     #######      ##      ####  ########
    */


    function getTasks() {
        var tasks = {
            dn: [],
            ndn: []
        };

        var lineCount = _editor.lineCount();

        for (var i = 0; i < lineCount; i++) {

            var line = _editor.getLine(i);

            if (line.match(/^- \[ \] .+/)) {
                tasks.ndn.push({
                    txt: line,
                    line: i
                });
            } else if (line.match(/^- \[x\] .+/)) {
                tasks.dn.push({
                    txt: line,
                    line: i
                });
            }
        }
        return tasks;
    }


    function getFileSizeInMB(path) {
        var stats = fs.statSync(path);
        var fileSizeInBytes = stats["size"];
        return (fileSizeInBytes / 1000000).toFixed(2);
    }





    /*
    ##     ##  ########   ########      ###
    ###   ###  ##            ##        ## ##
    #### ####  ##            ##       ##   ##
    ## ### ##  ######        ##      ##     ##
    ##     ##  ##            ##      #########
    ##     ##  ##            ##      ##     ##
    ##     ##  ########      ##      ##     ##
    */
    function getMetaFields() {
        var metaFieldsObject = {};
        metaFieldsObject.endPoint = 0;
        var yamlStart = false;

        var lineCount = (_editor.lineCount() < 25) ? _editor.lineCount() : 25;


        for (var i = 0; i < lineCount; i++) {

            var line = _editor.getLine(i);

            if (i == lineCount - 1) {
                line = line + "\n";
            }


            var lineLength = line.length;

            // support for YAML start
            if (i == 0 && line == "---") {
                metaFieldsObject.endPoint += lineLength + 1;
                yamlStart = true;
                continue;
            }

            var match = line.match(/^\s*[\w-]+:.*$/);


            if (match) {

                metaFieldsObject.endPoint += lineLength + 1;

                var spl = line.split(":")
                var key = spl[0].trim().toLowerCase();
                var value = spl[1].trim();

                // handle tags
                if (key == "tags" || key == "keywords") {
                    key = "tags";
                    var tags = [];
                    tags = value.replace(/rainbow/gi, "red orange yellow green blue indigo violet").split(/\s+|\s*,\s*/);
                    value = tags;
                }
                // handle due
                else if (key == "due") {
                    // check if valid date
                    var d = Date.parse(value);
                    value = (!isNaN(d)) ? d : 0;
                    // console.log(value, moment(value).format());
                }

                // handle goal
                else if (key == "goal") {
                    if (/.*c.*/.test(value)) {
                        value = value.replace(/c/i, "");
                        metaFieldsObject['goalUnit'] = "c";
                    } else if (/.*w.*/.test(value)) {
                        value = value.replace(/w/i, "");
                        metaFieldsObject['goalUnit'] = "w";
                    } else {
                        metaFieldsObject['goalUnit'] = "w";
                    }

                    value = value.replace(/[,\.]/g, ""); // remove comas and fullstops
                    value = (isNaN(parseInt(value, 10))) ? 0 : parseInt(value, 10);

                }

                metaFieldsObject[key] = value;

            } else if (line == "..." || line == "---") { // support for YAML end
                metaFieldsObject.endPoint += lineLength + 1;
                return metaFieldsObject;
            }

            else { // break and return if no more meta data lines
                return metaFieldsObject;
            }
        }
    }

    function getDocumentName() {
        var lineCount = (_editor.lineCount() < 50) ? _editor.lineCount() : 50;

        var documentName = "";
        var firstValid = "";
        var isHeading = false;


        // first check if there is explicit name in meta data
        var md = getMetaFields();
        if (md && md.title) return md.title;
        if (md && md.name) return md.name;

        for (var i = 0; i < lineCount; i++) {
            var line = _editor.getLine(i).trim();

            // skip if line is empty or comment or meta line
            if (line == "" || line.startsWith("//") || line.match(/^[\w-]+:.*$/)) {
                continue;
            }
            // set first valid line if not yet set for fallback to generated name
            firstValid = (firstValid == "") ? line : firstValid;

            // check for markdovn header
            if (line.startsWith("#")) {
                documentName = line.replace(/^#+/, "");
                isHeading = true;
                break;
            }
        }

        firstValid = (firstValid == "") ? "SCRATCH" : firstValid;
        documentName = (documentName == "") ? firstValid : documentName;
        documentName = documentName.replace(/\s/g, " ");
        documentName = (documentName.length > 150) ? documentName.substring(0, 150) + "..." : documentName; // max name length

        return documentName.trim();
    }






    function getDocumentsPaletteItems() {

        // $("#filterdoc_list input").val(""); $("#doc_list").jqxListBox("refresh");

        var itemsA = _currentDocList;
        var paletteA = [];

        for (var i = 0; i < itemsA.length; i++) {

            var ob = itemsA[i];

            paletteA.push({
                id: '',
                string: ob.nam.substring(0, 60),
                shortcut: moment(ob.dtm).fromNow(true),
                // action is wrong!!!!
                action: '$("#doc_list").jqxListBox("clearFilter"); $("#doc_list").jqxListBox("selectItem", $("#doc_list").jqxListBox("getItem",' + i.toString() + ')); $("#doc_list").jqxListBox("invalidate");',
                // action: 'if ($("#filterdoc_list input").val() != "") { $("#filterdoc_list input").val(""); $("#doc_list").jqxListBox("refresh"); }; $("#doc_list").jqxListBox("selectItem", $("#doc_list").jqxListBox("getItem",' + i.toString() + ')); $("#doc_list").jqxListBox("invalidate");',
                type: "eval",
                sublist: [],
                persist: false,
                icon: '&#xf0f6;',
                widget: {
                    type: "eval",
                    value: '"' + stats.add000separator(ob.wrd) + ' wrd"'
                }
            });
        }

        return paletteA;
    }




    /*
    ########      ###      ######     ######
       ##        ## ##    ##    ##   ##    ##
       ##       ##   ##   ##         ##
       ##      ##     ##  ##   ####   ######
       ##      #########  ##    ##         ##
       ##      ##     ##  ##    ##   ##    ##
       ##      ##     ##   ######     ######
    */
    // function tagsStringFromArray(ta) {
    // 	ta = (Array.isArray(ta)) ? ta : [];
    // 	var withHash = ta.map(function(item) {
    // 		return "#" + item;
    // 	});

    // 	return tagsString = withHash.join(" ");
    // }


    function getDistinctTags(callback) {
        var allTags = [];
        collection
            .find({
                del: false,
                arc: false
            }, {
                met: 1
            })
            .toArray(function (err, arr) {
                // assert.equal(null, err);
                for (var i = 0; i < arr.length; i++) {
                    allTags = allTags.concat(arr[i].met.tags);
                }
                allTags = allTags.distinctByCount().map(function (item) {
                    return "#" + item
                });
                callback(allTags);
            });
    }


    // Call constructor
    __init();




    /*
       ###     ########   ####
      ## ##    ##     ##   ##
     ##   ##   ##     ##   ##
    ##     ##  ########    ##
    #########  ##          ##
    ##     ##  ##          ##
    ##     ##  ##         ####
    */
    return {
        getDb: (function () {
            return db;
        }),
        getCollection: (function () {
            return collection;
        }),
        getCurrentId: (function () {
            return _currentId;
        }),
        newProject: newProject,
        getProjectsArr: getProjectsArr,
        showProjectsPalette: showProjectsPalette,
        setProjectsTree: setProjectsTree,
        setProject: setProject,
        closeProject: closeProject,
        deleteProject: deleteProject,
        moveProjectToTop: moveProjectToTop,
        renameProject: renameProject,
        assignDocumentToProject: assignDocumentToProject,
        copyProjectTxt: copyProjectTxt,
        getCombinedProject: getCombinedProject,
        getLastSearchesArray: getLastSearchesArray,
        getLastSearchesArray: getLastSearchesArray,
        doFullTextSearch: doFullTextSearch,
        getDocumentName: getDocumentName,
        uiUpdateDocumentName: uiUpdateDocumentName,
        getDistinctTags: getDistinctTags,
        getMetaFields: getMetaFields,
        getDocumentObject: getDocumentObject,
        newDocument: newDocument,
        loadDocument: loadDocument,
        isCleanMarks: isCleanMarks,
        deleteDocument: deleteDocument,
        removeDocument: removeDocument,
        unDeleteDocument: unDeleteDocument,
        emptyTrash: emptyTrash,
        updateFields: updateFields,
        updateDocument: updateDocument,
        setDocumentList: setDocumentList,
        compactCollection: compactCollection,
        backupDatabase: backupDatabase,
        openDBfolder: openDBfolder,
        openBackupFolder: openBackupFolder,
        openSettingFiles: openSettingFiles,
        openDefaultThemeFile: openDefaultThemeFile,
        closeDB: closeDB,
        showDocumentHistoryPalette: showDocumentHistoryPalette,
        restoreDocumentFromHistory: restoreDocumentFromHistory,
        getDocumentsPaletteItems: getDocumentsPaletteItems
    };

})();