/*
 * writemonkey.js core
 * This handles all core app things.
 */

var gui = require('nw.gui');
var fs = require('fs');
var app = gui.App;
var os = require('os');
var win = gui.Window.get();
var path = require('path');
var args = window.gui.App.argv;
var child = require('child_process');
var _s = require("underscore.string");
var Db = require('tingodb')({searchInArray:true}).Db;
// var assert = require('assert');
var opener = require('opener');
var util = require('util');
var vm = require('vm');
var crp = require('crypto');
var nodePandoc = require('node-pandoc');

// var nspell = require('nspell');


var mdIt = require('markdown-it')({
	html: true,
	breaks: true,
	typographer: true,
	quotes: '“”‘’'

});
mdIt.use(require('markdown-it-footnote'));
mdIt.use(require('markdown-it-mark'));
mdIt.use(require('markdown-it-toc'));

// WM GLOBALS
var _strings = {};
var _settings = {};
var _persistent = {};
var _plugins = {};
var _modules = {};
var _domElements = {};
var _editor = null;

var _currentDoc = null;
var _currentDocList = [];
var _currentJumpsList = [];
var _currentRepoList = [];
var _recentsBuffer = {};
var _currentId = -1;
var _isReadOnlyMarker = null;
var _currentProjectId = -1;
var _currentProjectsList = [];
var _currentWordCount = 0;
var _lastChangePosition = null;
var _currentTheme = "";
var _isAppClosing = false;
var _isEditorSaved = true;
var _foldRangeFinder = new CodeMirror.fold.combine(CodeMirror.fold.indent, CodeMirror.fold.markdown);
var _anchoredReading = {lastPos: -1, lastBackup: -1, skippedWords: 0};

// SPELLING GLOBALS
var num_loaded = 0;
var aff_data = "";
var dic_data = "";
var typo;
//var mySpell;
var suggestContextMenu;
var loadedDictionaries = {};
var ignoreList = [];





var wm = (function(window) {


	/*
	####  ##    ##   ####  ########
	 ##   ###   ##    ##      ##
	 ##   ####  ##    ##      ##
	 ##   ## ## ##    ##      ##
	 ##   ##  ####    ##      ##
	 ##   ##   ###    ##      ##
	####  ##    ##   ####     ##
	*/

	function __init() {
		loadStrings();
		loadSettings();
		setCM();
	}



	function initUI() {

		if (!localStorage["APP_INIT"]) {
			localStorage["VIEWPORT_SOURCES_PANE_SIZE"] = 0;
			localStorage["VIEWPORT_SOURCES_PANE_VISIBLE"] = 1;
			localStorage["VIEWPORT_DOCUMENT_PANE_COMPACT"] = 1;
			localStorage["VIEW_TYPEWRITER_SCROLLING"] = 1;
			localStorage['VIEWPORT_AUTO_FADE_ENABLED'] = 0;
			localStorage["APP_INIT"] = 1;
			localStorage["SOUNDS_WHITE_NOISE_SELECTED_SCHEME"] = "";
			localStorage['VIEWPORT_UI_ZOOM'] = 100;
			localStorage['PROJECTS_TREE_EXPANDED_DATA'] = "{}";
			localStorage["STAT_IN_INFOBAR"] = "getWords";
			localStorage["VIEWPORT_MAIN_FONT"] = "Hack Mono";
			localStorage["VIEWPORT_MAIN_FONT_LINE_SPACING"] = 1.5;
		}

		// Set window position
		if (localStorage["WINDOW_STARTUP_PARAMS"]) {
			var winParams = JSON.parse(localStorage["WINDOW_STARTUP_PARAMS"]);
			if (winParams && winParams.x != winParams.y) {
				win.width = winParams.w;
				win.height = winParams.h;
				win.x = winParams.x;
				win.y = winParams.y;
			}
		}


		// Set lock on top to false
		localStorage['VIEWPORT_WINDOW_LOCKED_ON_TOP'] = 0;




		// set theme
		setTheme();

		if (_settings.start_in_full_screen) win.enterFullscreen();
		setAutohideTimer();
		// if (_settings.session_inactivity_signal) setSessionSignal();

		// show / hide panes
		setPanes();

		// set initial zoom
		_editor.execCommand("initZoom");


		var uiZoom = localStorage['VIEWPORT_UI_ZOOM'] ? localStorage['VIEWPORT_UI_ZOOM'] : 100;
		$("html").css({"font-size": uiZoom.toString() + "%"});

		// apply ui filter
		applyFilters();

		// adjust cm height
		$(".editor-placeholder").height($("#grid").height() - 40);

		// disable markdown code block (indented first line)
		if (_settings.normal_line_indenting) mdIt.disable('code');


		console.log("ALL SYSTEMS GO!");

	}

	function saveUI() {}


	/*
	 ######    ########   ########   ########   ####  ##    ##    ######     ######
	##    ##   ##            ##         ##       ##   ###   ##   ##    ##   ##    ##
	##         ##            ##         ##       ##   ####  ##   ##         ##
	 ######    ######        ##         ##       ##   ## ## ##   ##   ####   ######
	      ##   ##            ##         ##       ##   ##  ####   ##    ##         ##
	##    ##   ##            ##         ##       ##   ##   ###   ##    ##   ##    ##
	 ######    ########      ##         ##      ####  ##    ##    ######     ######
	*/

	//console.log(getJsonFromFile("settings/settings_main.json"));


	function loadSettings() {

		var defaults = getObjFromJsonFile("settings/settings_main.json");
		var user = getObjFromJsonFile("settings/settings_main_user.json");

		// map user settings to defaults
		for (key in user) {
			defaults[key] = user[key];
		}

		_settings = defaults;
		console.log("Settings loaded");
	}



	/*
	 ######    ########   ########   ####  ##    ##    ######     ######
	##    ##      ##      ##     ##   ##   ###   ##   ##    ##   ##    ##
	##            ##      ##     ##   ##   ####  ##   ##         ##
	 ######       ##      ########    ##   ## ## ##   ##   ####   ######
	      ##      ##      ##   ##     ##   ##  ####   ##    ##         ##
	##    ##      ##      ##    ##    ##   ##   ###   ##    ##   ##    ##
	 ######       ##      ##     ##  ####  ##    ##    ######     ######
	*/

	function loadStrings() {
		var defaults = getObjFromJsonFile("lang/strings_defaults.json");
		_strings = defaults;
		console.log("Strings loaded");
	}



	/*
	      ##    ######     #######   ##    ##
	      ##   ##    ##   ##     ##  ###   ##
	      ##   ##         ##     ##  ####  ##
	      ##    ######    ##     ##  ## ## ##
	##    ##         ##   ##     ##  ##  ####
	##    ##   ##    ##   ##     ##  ##   ###
	 ######     ######     #######   ##    ##
	*/
	function getObjFromJsonFile(path) {
		var json, obj, key, fileFound = true;
		try {
			json = fs.readFileSync(path, {
				encoding: "utf-8"
			});
		} catch (err) {
			handleError("File " + path + " not found!");
			fileFound = false;
		}
		if (fileFound == true) {
			try {
				obj = JSON.parse(minifyJson(json));
			} catch (err) {
				handleError("Invalid JSON file: " + path);
			}
			return obj;
		}
	}


	function minifyJson(a) {
		var f, g, k, l, b = /"|(\/\*)|(\*\/)|(\/\/)|\n|\r/g,
			c = !1,
			d = !1,
			e = !1,
			h = [],
			i = 0,
			j = 0;

		for (b.lastIndex = 0; f = b.exec(a);) k = RegExp.leftContext, l = RegExp.rightContext,
			d || e || (g = k.substring(j), c || (g = g.replace(/(\n|\r|\s)*/g, "")), h[i++] = g),
			j = b.lastIndex, '"' != f[0] || d || e ? "/*" != f[0] || c || d || e ? "*/" != f[0] || c || !d || e ?
			"//" != f[0] || c || d || e ? "\n" != f[0] && "\r" != f[0] || c || d || !e ? d || e || /\n|\r|\s/
			.test(f[0]) || (h[i++] = f[0]) : e = !1 : e = !0 : d = !1 : d = !0 :
			(g = k.match(/(\\)*$/), c && g && 0 != g[0].length % 2 || (c = !c), j--, l = a.substring(j));

		return h[i++] = l, h.join("");
	}



	/*
	 ######    ##     ##  ####  ########   ########    #######   ########
	##    ##   ###   ###   ##   ##     ##  ##     ##  ##     ##  ##     ##
	##         #### ####   ##   ##     ##  ##     ##  ##     ##  ##     ##
	##         ## ### ##   ##   ########   ########   ##     ##  ########
	##         ##     ##   ##   ##   ##    ##   ##    ##     ##  ##   ##
	##    ##   ##     ##   ##   ##    ##   ##    ##   ##     ##  ##    ##
	 ######    ##     ##  ####  ##     ##  ##     ##   #######   ##     ##
	*/



	function setCM() {
		// Codemirror default options
		var opt = {
			readOnly: 'nocursor', // disable this on start up
			dragDrop: true,
			flattenSpans: true,
			gutters: ["CodeMirror-universalgutter", "CodeMirror-foldgutter"],
			foldGutter: {
                rangeFinder: _foldRangeFinder
            },
			foldOptions: {
				widget: "\u2026",
				scanUp: false,
				rangeFinder: _foldRangeFinder
			},
			fixedGutter: true,
			indentUnit: _settings.tabs_number_of_spaces,
			tabSize: _settings.tabs_number_of_spaces,
			indentWithTabs: _settings.indent_with_tabs,
			inputStyle: "textarea",
			scrollbarStyle: "overlay",
			showWhitespaces: false,
			showCursorWhenSelecting: false,
			styleActiveLine: localStorage["VIEW_STYLE_ACTIVE_LINE"] && localStorage["VIEW_STYLE_ACTIVE_LINE"] == "1" ? true : false,
			fadeInactiveParagraps: localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] && localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] == "1" ? true : false,

			styleCurrentSentence: localStorage["VIEW_STYLE_CURRENT_SENTENCE"] && localStorage["VIEW_STYLE_CURRENT_SENTENCE"] == "1" ? true : false,

			styleSelectedText: true,
			selectionPointer: true,
			lineNumbers: false,
			lineWrapping: true,

			autoCloseBrackets: {
				pairs: _settings.smart_characters,
				triples: "",
    			explode: "[]{}",
				override: true
			},

			maxHighlightLength: 10000, // def is 10000
			mode: "gfm", //spell-checker

			theme: "wm-default",
			keyMap: "sublime",
			// extraKeys: {

			// },
			typewriterScrolling: localStorage.getItem("VIEW_TYPEWRITER_SCROLLING") === null || localStorage["VIEW_TYPEWRITER_SCROLLING"] == "1" ? true : false,
			highlightComments: true,
			highlightMarked: true,
			highlightFootnotes: true,
			// rulers: [{color: "#fff", column: 0, lineStyle: "dashed"}, {color: "#fff", column: 80, lineStyle: "dashed"}],
			scrollPastEnd: true,
			electricChars: false,
			autofocus: true,
			cursorHeight: 1,
			singleCursorHeightPerLine: true,
			cursorScrollMargin: 0,
			allowDropFileTypes: _settings.files_allow_drop_types,
			showSentenceBreakup: localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] && localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] == "1" ? true : false,
			showAnchoredReading: localStorage["VIEW_SHOW_ANCHORED_READING"] && localStorage["VIEW_SHOW_ANCHORED_READING"] == "1" ? true : false

		};

		$("#grid").html('<div id="editor-placeholder" class="editor-placeholder"><textarea id="cdmeditor"></textarea></div>');
		// Create cm instance in global space
		_editor = CodeMirror.fromTextArea(document.getElementById("cdmeditor"), opt);

	}




	/*
	##    ##    #######   ########   ####   ######    ########    ######
	###   ##   ##     ##     ##       ##   ##    ##   ##         ##    ##
	####  ##   ##     ##     ##       ##   ##         ##         ##
	## ## ##   ##     ##     ##       ##   ##         ######      ######
	##  ####   ##     ##     ##       ##   ##         ##               ##
	##   ###   ##     ##     ##       ##   ##    ##   ##         ##    ##
	##    ##    #######      ##      ####   ######    ########    ######
	*/
	function showNotice(noticeText, options) {

		var defaultOptions = {
			autoClose: true,
			autoCloseDelay: 6000,
			position: "bottom-right",
			blink: false,
			width: 14,
			click: function(){}
		};

		if (typeof options == 'object') {
			options = $.extend(defaultOptions, options);
		} else {
			options = defaultOptions;
		}


		var uniqueId = "wm_" + getNewUniqueID();
		$('<div/>', {
			id: uniqueId,
			html: '<div style="overflow-wrap: break-word; word-wrap: break-word; font-size: .8rem; width: ' + (options.width - 1).toString() + 'rem;">' +  noticeText.replace(/\n/g, "<br>") + '</div>',
			class: "wm-notice",
			style: ''


		}).appendTo('body');

		$("#" + uniqueId).jqxNotification({
			// appendContainer: "#main_pane",
			browserBoundsOffset: 40,
			width: options.width.toString() + 'rem',
			position: options.position,
			opacity: 1,
			autoOpen: false,
			animationCloseDelay: 0,
			blink: options.blink,
			showCloseButton: false,
			closeOnClick: true,
			autoClose: options.autoClose,
			autoCloseDelay: options.autoCloseDelay,
			template: null,
			theme: "metro"

		}).on("click", options.click);

		$("#" + uniqueId).jqxNotification("open");

	}



	function updateIconBadge(text) {
		win.setBadgeLabel(text);
	}


	/*
	########   ########   ########    #######   ########    ######
	##         ##     ##  ##     ##  ##     ##  ##     ##  ##    ##
	##         ##     ##  ##     ##  ##     ##  ##     ##  ##
	######     ########   ########   ##     ##  ########    ######
	##         ##   ##    ##   ##    ##     ##  ##   ##          ##
	##         ##    ##   ##    ##   ##     ##  ##    ##   ##    ##
	########   ##     ##  ##     ##   #######   ##     ##   ######
	*/

	function handleError(error) {
		console.log(error);
	}




	/*
	 #######   ##    ##   ##          #######      ###     ########
	##     ##  ###   ##   ##         ##     ##    ## ##    ##     ##
	##     ##  ####  ##   ##         ##     ##   ##   ##   ##     ##
	##     ##  ## ## ##   ##         ##     ##  ##     ##  ##     ##
	##     ##  ##  ####   ##         ##     ##  #########  ##     ##
	##     ##  ##   ###   ##         ##     ##  ##     ##  ##     ##
	 #######   ##    ##   ########    #######   ##     ##  ########
	*/
	win.on('loaded', function() {
		initUI();
		win.show();
		win.focus();
		setTimeout(function() {
			win.show();
		}, 5000);
	});

	// check if backup is necesary on blur
	win.on('blur', function() {
		setTimeout(function() {
			if (_settings.backup_database_automatically == 0 || this._isAppClosing == true)
				return;

				var last = localStorage['DATABASE_LAST_BACKUP'] ? localStorage['DATABASE_LAST_BACKUP'] : 0;
			var period = moment().subtract(_settings.backup_database_automatically, 'hours').format("x");

			if (last == 0 || last < period) {
				// do backup
				storage.backupDatabase(true);
			}
		}, 500);
	});


	/*
	 #######   ##    ##    ######    ##          #######    ######    ########
	##     ##  ###   ##   ##    ##   ##         ##     ##  ##    ##   ##
	##     ##  ####  ##   ##         ##         ##     ##  ##         ##
	##     ##  ## ## ##   ##         ##         ##     ##   ######    ######
	##     ##  ##  ####   ##         ##         ##     ##        ##   ##
	##     ##  ##   ###   ##    ##   ##         ##     ##  ##    ##   ##
	 #######   ##    ##    ######    ########    #######    ######    ########
	*/
	win.on('close', function() {

		this.hide();
		this._isAppClosing = true;

		// store window size and position

		if (!win.isFullscreen && win.x != win.y) { // only save if not full screen or minimized
			localStorage["WINDOW_STARTUP_PARAMS"] = JSON.stringify({
				"x": win.x,
				"y": win.y,
				"w": win.width,
				"h": win.height
			});
		}

		// count words
		storage.getCollection()
			.find({
				del: false,
				arc: false
			}, {
				wrd: 1
			})
			.toArray(function (err, arr) {
				var sum = 0;
				for (var i = 0; i < arr.length; i++) {
					sum += arr[i].wrd;
				}
				localStorage["STATS_LIBRARY_WORD_COUNT"] = sum;
				console.log(arr);

				// now save / close wm

				if (_editor.isClean() && storage.isCleanMarks()) {
					saveUI();
					storage.closeDB();
					app.quit();
				}

				var docO = storage.getDocumentObject();

				// update file if needed
				if (docO.pth && docO.pth.length > 0) {
					files.saveWmFile(docO.pth, docO.txt);
				}

				storage.getCollection().update({
					_id: storage.getCurrentId()
				}, docO, function (err, result) {
					saveUI();
					storage.closeDB();
					app.quit();
				});

			}); // end collection find
	});




	/*
	####  ##    ##   ########    #######   ########      ###     ########
	 ##   ###   ##   ##         ##     ##  ##     ##    ## ##    ##     ##
	 ##   ####  ##   ##         ##     ##  ##     ##   ##   ##   ##     ##
	 ##   ## ## ##   ######     ##     ##  ########   ##     ##  ########
	 ##   ##  ####   ##         ##     ##  ##     ##  #########  ##   ##
	 ##   ##   ###   ##         ##     ##  ##     ##  ##     ##  ##    ##
	####  ##    ##   ##          #######   ########   ##     ##  ##     ##
	*/

	// init new worker
	var statWorker = new Worker("core/stats_count_worker.js");


	function startNewSession() {

		_currentDoc.ses = {
			startTime: new Date().getTime(),
			startCount: stats.getWords(_editor.getValue()).count,
			countUnit: "words"
		};

		// add session start event
		var newSessionE = new CustomEvent("newSessionE", {
			detail: {
				sesInfo: _currentDoc.ses
			}
		});
		document.body.dispatchEvent(newSessionE);

		// resetSessionSignalTimeout();
		showNotice("New session started!");
	}




	function pauseUnpauseCurrentSession() {
		_currentDoc.ses.isPaused = true;
		_currentDoc.ses.pauseSeconds = 0;
		showNotice("Session paused. Click to resume!", {autoClose: false});
	}



	function stopCurrentSession() {

		_currentDoc.ses = {};

		// add session stop event
		var stopSessionE = new CustomEvent("stopSessionE", {
			detail: {
				sesInfo: _currentDoc.ses
			}
		});
		document.body.dispatchEvent(stopSessionE);
		showNotice("Session stopped!");

	}



	// when message from doCount is received
	statWorker.onmessage = function(e) {

		// do nothing if infobar is hidden
		if (!window.document.hasFocus()) return; // !isInfoBarVisible ||



		var m = e.data.prgString;
		_currentWordCount = e.data.words;


		$("#info_bar_right").html('<span id="info_bar_stats">' + m + '</span>');

		updateSessionInfo();


		// set progress bar if goal is set
		// else hide progress bar
		if (_currentDoc &&_currentDoc.met && _currentDoc.met.goal) {

			// Show progress bar
			// $("#info_bar_bottom").css("flex", "0 28px");

			$("#info_bar_bottom").css("flex", "0 1.8rem");
			// $("#info_bar_bottom").css("flex", "0 " + vSize.toString() + "px");


			$("#info_bar_progressbar_outer").css("height", ".2rem");

			var percent = (e.data.words / _currentDoc.met.goal * 100).toFixed(3);

			var previousSessionWords = (_currentDoc.ses && _currentDoc.ses.startCount) ? _currentDoc.ses.startCount : 0;

			var maxWords = (e.data.words > _currentDoc.met.goal) ? _currentDoc.met.goal : e.data.words;
			var sessionPercent = ((maxWords - previousSessionWords) / e.data.words * 100).toFixed(3);
			sessionPercent = (previousSessionWords <= 0) ? 0 : sessionPercent;


			// add last session part
			if (percent <= 100) {
				$("#info_bar_progressbar_session").css({
					"width": sessionPercent.toString() + "%",
					"float": "right"
				});
			}


			// if more than 100% - add overflow bar
			if (percent > 100) {
				var overflow = (percent - 100 > 100) ? 100 : percent - 100;
				percent = 100; // set pecent to 100 as it can't be higher
				$("#info_bar_progressbar_overflow").css("width", overflow + "%");
			} else {
				$("#info_bar_progressbar_overflow").css("width", "0%");
			}


			// show total percent
			$("#info_bar_progressbar_inner").css("width", percent + "%");
			// $("#info_bar_progressbar_inner").css("opacity", ".7");


			// show due today goal as dot / same color as #info_bar_progressbar_session
			// if (_currentDoc.met && _currentDoc.met.due) {
			// }



		} else { // hide progress bar

			$("#info_bar_progressbar_outer").css("height", "0px");
			$("#info_bar_bottom").css("flex", "0 1.4rem");
			$("#info_bar_bottom").css("padding-bottom", "0.2rem");

			$("#info_bar_progressbar_inner").css("width", "0%");
			$("#info_bar_progressbar_overflow").css("width", "0%");
			//$("#info_bar_progressbar_dot1").css("height", "0px");
		}

		var statToShow = ["getWords"];
		if (localStorage["STAT_IN_INFOBAR"] && localStorage["STAT_IN_INFOBAR"] != "getWords") {
			statToShow.push(localStorage["STAT_IN_INFOBAR"]);
		}


		// send new text to worker
		var messageIn = {
			text: _editor.somethingSelected() ? _editor.getSelection() : _editor.getValue(),
			units: statToShow,
			readingWPM: _settings.reading_wpm,
			speakingWPM: _settings.speaking_wpm,
			standard_pages: _settings.standard_pages
		};


		// statWorker.postMessage(messageIn);
		messageToStatsWorker(messageIn);

	};


	function messageToStatsWorker(msg) {
		statWorker.postMessage(msg);
	}



	function animateProgressBar() {
		$("#info_bar_progressbar_inner").hide().delay(1500).show("slow");
		$("#info_bar_progressbar_overflow").hide().delay(1500).show("slow");
	}


	// Set date / time
	var clockTimer = setInterval(function() {
		var nowString = moment().format(_settings.infobar_datetime_format);
		$("#info_bar_left").html('<span id="info_bar_datetime">' + nowString + '</span>');
	}, 2000);





	function updateSessionInfo() {

		var sessionString = "";

		if (_currentDoc != null && _currentDoc.ses != null && _currentDoc.ses.startCount != null) {

			var now = new Date().getTime();
			var sessCount = _currentWordCount - _currentDoc.ses.startCount;
			var sessionStart = _currentDoc.ses.startTime;



			var h = moment(now).diff(sessionStart, "hours");
			var m = moment(now).diff(sessionStart, "minutes") - (h * 60);
			m = m < 10 ? '0' + m : m;

			// if (h < 12) {
			var wpm = (sessCount / moment(now).diff(sessionStart, "seconds") * 60).toFixed(0);
			wpm = wpm < 0 || isNaN(wpm) ? 0 : wpm;

			sessionString = 'w ' + stats.add000separator(sessCount);

			sessionString += '&nbsp;&nbsp;&#8226;&nbsp;&nbsp;t ' + h + ':' + m;
			sessionString += '&nbsp;&nbsp;&#8226;&nbsp;&nbsp;wpm ' + Math.abs(wpm).toString();
			// }
		}
		$("#info_bar_top_right").html('<span id="info_bar_session_info">' + sessionString + '</span>');
	}



	/*
	   ###     ##     ##  ########    #######   ########      ###     ########   ########
	  ## ##    ##     ##     ##      ##     ##  ##           ## ##    ##     ##  ##
	 ##   ##   ##     ##     ##      ##     ##  ##          ##   ##   ##     ##  ##
	##     ##  ##     ##     ##      ##     ##  ######     ##     ##  ##     ##  ######
	#########  ##     ##     ##      ##     ##  ##         #########  ##     ##  ##
	##     ##  ##     ##     ##      ##     ##  ##         ##     ##  ##     ##  ##
	##     ##   #######      ##       #######   ##         ##     ##  ########   ########
	*/


	var uiAutohideTimer = null;
	var isUIHidden = false;
	var numberOfChanges = 0;

	function setAutohideTimer() { // must be called only once at init!!!!!!!!



		_editor.on("change", function(cm, e) {
			if (localStorage['VIEWPORT_AUTO_FADE_ENABLED'] == 0) return;

			clearTimeout(uiAutohideTimer);
			uiAutohideTimer = setTimeout(unfadeUI, _settings.ui_autofadein_delay);

			if (e.origin == "+input" && e.text.length == 1) {
				numberOfChanges++;
			} else {
				numberOfChanges = 0;
			}

			if (!isUIHidden && numberOfChanges >= _settings.ui_autofadeout_actions_before_trigger) {
				fadeUI();
			}

		});

		_editor.on("blur", function() {// reset counter to zero when focus is lost
			if (isUIHidden) {
				unfadeUI(1);
			}
		});


		_editor.on("swapDoc", function(cm) {
			if (isUIHidden) {
				unfadeUI(1);
			}
		});
	}

	function fadeUI() {
		isUIHidden = true;

		if (_settings.ui_autofade_include_infobar == true) {
			$("#info_bar_table_top").fadeTo(_settings.ui_autofadeout_animation_time, _settings.ui_autofade_amount_infobar);
			$("#info_bar_bottom").fadeTo(_settings.ui_autofadeout_animation_time, _settings.ui_autofade_amount_infobar);
		}

		var upr = $("#util_pane_right");
		var upl = $("#util_pane_left");
		var mp = $("#module_pane");

		if (localStorage['VIEWPORT_DOCUMENT_PANE_VISIBLE'] == 1 && upr.is(':visible')) upr.fadeTo(_settings.ui_autofadeout_animation_time, _settings.ui_autofade_amount_panes);
		if (localStorage['VIEWPORT_SOURCES_PANE_VISIBLE'] == 1 && upl.is(':visible')) upl.fadeTo(_settings.ui_autofadeout_animation_time, _settings.ui_autofade_amount_panes);
		if (localStorage['VIEWPORT_MODULE_PANE_VISIBLE'] == 1 && mp.is(':visible')) mp.fadeTo(_settings.ui_autofadeout_animation_time, _settings.ui_autofade_amount_panes);
	}


	function unfadeUI(forcedAnimTime) {
		var animTime = forcedAnimTime ? forcedAnimTime : _settings.ui_autofadein_animation_time;
		isUIHidden = false;
		numberOfChanges = 0;

		if (_settings.ui_autofade_include_infobar == true) {
			$("#info_bar_table_top").fadeTo(animTime, 1);
			$("#info_bar_bottom").fadeTo(animTime, 1);
		}


		var upr = $("#util_pane_right");
		var upl = $("#util_pane_left");
		var mp = $("#module_pane");

		if (localStorage['VIEWPORT_DOCUMENT_PANE_VISIBLE'] == 1 && upr.is(':visible')) upr.fadeTo(animTime, 1);
		if (localStorage['VIEWPORT_SOURCES_PANE_VISIBLE'] == 1 && upl.is(':visible')) upl.fadeTo(animTime, 1);
		if (localStorage['VIEWPORT_MODULE_PANE_VISIBLE'] == 1 && mp.is(':visible')) mp.fadeTo(animTime, 1);
	}


	function toggleAutoHide() {
		if (localStorage['VIEWPORT_AUTO_FADE_ENABLED'] && localStorage['VIEWPORT_AUTO_FADE_ENABLED'] == 1) {
			localStorage['VIEWPORT_AUTO_FADE_ENABLED'] = 0;
			showNotice("Distraction free fading is disabled!");

		} else {
			localStorage['VIEWPORT_AUTO_FADE_ENABLED'] = 1;
			showNotice("Distraction free fading is enabled!");
		}
	}


	/*
	########      ###     ##    ##   ########    ######
	##     ##    ## ##    ###   ##   ##         ##    ##
	##     ##   ##   ##   ####  ##   ##         ##
	########   ##     ##  ## ## ##   ######      ######
	##         #########  ##  ####   ##               ##
	##         ##     ##  ##   ###   ##         ##    ##
	##         ##     ##  ##    ##   ########    ######
	*/

	function isWmFullScreen() {
		return (win.isFullscreen && localStorage.VIEWPORT_SOURCES_PANE_VISIBLE == 0 && localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE == 0 && localStorage.VIEWPORT_MODULE_PANE_VISIBLE == 0)
	}

	function toggleLockOnTop() {

		if (localStorage['VIEWPORT_WINDOW_LOCKED_ON_TOP'] == 1) {
			win.setAlwaysOnTop(false);
			localStorage['VIEWPORT_WINDOW_LOCKED_ON_TOP'] = 0;
			showNotice("No longer locked on top of all other windows");
		} else {
			win.setAlwaysOnTop(true);
			localStorage['VIEWPORT_WINDOW_LOCKED_ON_TOP'] = 1;
			showNotice("Locked on top of all other windows");
		}
	}


	function toggleDistractionFree() {
		win.toggleFullscreen();
		_editor.execCommand("resetZoom");
		if (win.isFullscreen) hidePanes(); else showPanes();
	}


	function setPanes() {

		$("#util_pane_right").css('flex', '0 0 ' + (localStorage['VIEWPORT_DOCUMENT_PANE_SIZE'] ? localStorage['VIEWPORT_DOCUMENT_PANE_SIZE'] : "200") + 'px');

		if (localStorage.VIEWPORT_SOURCES_PANE_SIZE == 0) {
			$("#util_pane_left").css("flex", "0 0 0px");
		} else {
			$("#util_pane_left").css('flex', '0 0 ' + (localStorage['PROJECTPANE_WIDTH'] ? localStorage['PROJECTPANE_WIDTH'] : "100") + 'px');
		}


		$("#util_pane_left").toggle(localStorage.VIEWPORT_SOURCES_PANE_VISIBLE == 0 ? false : true);

		$("#toggle_sources").html($("#util_pane_left").width() == 0 ? "&#xf138;" : "&#xf137;");


		$("#util_pane_right").toggle(localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE == 0 ? false : true);
		$("#module_pane").toggle(localStorage.VIEWPORT_MODULE_PANE_VISIBLE == 0 ? false : true);




		_editor.focus(); _editor.refresh();
	}

	function hidePanes() {

		localStorage.VIEWPORT_SOURCES_PANE_VISIBLE = 0;
		localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE = 0;
		localStorage.VIEWPORT_MODULE_PANE_VISIBLE = 0;

		$("#util_pane_left").hide();
		$("#util_pane_right").hide();
		$("#module_pane").hide();
	}

	function showPanes() {

		localStorage.VIEWPORT_SOURCES_PANE_VISIBLE = 1;
		localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE = 1;
		localStorage.VIEWPORT_MODULE_PANE_VISIBLE = 1;

		$("#util_pane_left").show();
		$("#util_pane_right").show();
		$("#module_pane").show();
		$("#doc_list").jqxListBox('invalidate');
		unfadeUI(1);
	}

	function togglePanes() {
		if ($("#util_pane_right").is(':visible') || $("#module_pane").is(':visible')) {
			hidePanes();
		} else {
			showPanes();
		}
		_editor.focus(); _editor.refresh();
	}


	function toggleDocsJumps() {

		function focusVisible() {
			if ($('#doc_list').css("display") == "block") {
				$('#doc_list').focus();
				$('#doc_search').css("display", "block");
				//$("#doc_list").jqxListBox("invalidate");
			} else {
				$('#jumps_list').focus();
				$('#doc_search').css("display", "none");
				//$("#jumps_list").jqxListBox("invalidate");
			}
		}

		// check if visible, if not show, focus
		if (!$("#util_pane_right").is(':visible')) {

			$("#util_pane_left").show();
			$("#util_pane_right").show();
			localStorage.VIEWPORT_SOURCES_PANE_VISIBLE = 1;
			localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE = 1;
			focusVisible();
			unfadeUI(1);

		} else {

			if (!$("#doc_list").is(":focus") && !$("#jumps_list").is(":focus")) {
				focusVisible();
			} else {
				if ($('#doc_list').css("display") == "block") {
					$("#toggle_doc_jumps").html("&#xf15b;");
					$('#doc_list').css("display", "none");
					$('#jumps_list').css("display", "block");
					jumps.updateListBox();
				} else {
					$("#toggle_doc_jumps").html("&#xf0ca;"); // &#xf14e;
					$('#doc_list').css("display", "block");
					$('#jumps_list').css("display", "none");
				}
				focusVisible();
			}
		}
	}

	function toggleLeftPanes() {
		$("#util_pane_left").toggle(0);
		$("#util_pane_right").toggle(0);

		localStorage.VIEWPORT_SOURCES_PANE_VISIBLE = $("#util_pane_left").is(':visible') ? 1 : 0;
		localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE = $("#util_pane_right").is(':visible') ? 1 : 0;

		redrawListDirty();

		_editor.focus(); _editor.refresh();
		unfadeUI(1);
	}


	// show/hide panes based on win width
	win.on('resize', function() {

		if (win.width < _settings.auto_hide_panes_at) {
			if (localStorage.VIEWPORT_SOURCES_PANE_VISIBLE == 1) $("#util_pane_left").toggle(false);
			if (localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE == 1) $("#util_pane_right").toggle(false);
			if (localStorage.VIEWPORT_MODULE_PANE_VISIBLE == 1) $("#module_pane").toggle(false);
		} else {
			if (localStorage.VIEWPORT_SOURCES_PANE_VISIBLE == 1) $("#util_pane_left").toggle(true);
			if (localStorage.VIEWPORT_DOCUMENT_PANE_VISIBLE == 1) $("#util_pane_right").toggle(true);
			if (localStorage.VIEWPORT_MODULE_PANE_VISIBLE == 1) $("#module_pane").toggle(true);
		}
		_editor.refresh();
	});


	win.on('focus', function() {
		redrawListDirty();
	});

	// dirty hack to redraw document and jumps list
	function redrawListDirty() {
		setTimeout(function() {

			$("#sources_list").jqxTree("width", "99%");
			$("#sources_list").jqxTree("width", "100%");

			if ($('#doc_list').css("display") == "block") {
				$("#doc_list").jqxListBox("width", "99%");
				$("#doc_list").jqxListBox("width", "100%");
				$("#doc_list").jqxListBox('ensureVisible', $("#doc_list").jqxListBox('getSelectedIndex'));
			} else if ($('#jumps_list').css("display") == "block") {
				$("#jumps_list").jqxListBox("width", "99%");
				$("#jumps_list").jqxListBox("width", "100%");
				$("#jumps_list").jqxListBox('ensureVisible', $("#jumps_list").jqxListBox('getSelectedIndex'));
			}

		}, 10);
	}



	$("#edge").click(function() {
		toggleLeftPanes();
	});

	$("#edge2").click(function() {
		$("#module_pane").toggle(0);
		localStorage.VIEWPORT_MODULE_PANE_VISIBLE = $("#module_pane").is(':visible') ? 1 : 0;
		_editor.focus(); _editor.refresh();
		unfadeUI(1);
	});


	$("#edge").mouseenter(function() {
		$("#edge").css('background', 'linear-gradient(rgba(128, 128, 128, .25), rgba(128, 128, 128, .25)');
	});
	$("#edge").mouseleave(function() {
		$("#edge").css("background", "none");
	});

	$("#edge2").mouseenter(function() {
		$("#edge2").css('background', 'linear-gradient(rgba(128, 128, 128, .25), rgba(128, 128, 128, .25)');
	});
	$("#edge2").mouseleave(function() {
		$("#edge2").css("background", "none");
	});



	$("#toggle_sources").click(function() {

		if ($("#util_pane_left").width() == 0) {
			var w = localStorage['PROJECTPANE_WIDTH'] ? localStorage['PROJECTPANE_WIDTH'] : "110";
			$("#util_pane_left").css("flex", "0 0 " + w + "px");
			$("#toggle_sources").html("&#xf137;");
		} else {
			$("#util_pane_left").css("flex", "0 0 0px");
			$("#toggle_sources").html("&#xf138;");

		}
		localStorage.VIEWPORT_SOURCES_PANE_SIZE = $("#util_pane_left").width();

		_editor.focus(); _editor.refresh();
	});


	$("#toggle_doc_jumps").click(function() {
		if ($('#doc_list').css("display") == "block") {
			$("#toggle_doc_jumps").html("&#xf15b;");
			$('#doc_list').css("display", "none");
			$('#doc_search').css("display", "none");
			$('#jumps_list').css("display", "block");
			jumps.updateListBox();
		} else {
			$("#toggle_doc_jumps").html("&#xf0ca;");
			$('#doc_list').css("display", "block");
			$('#doc_search').css("display", "block");
			$('#jumps_list').css("display", "none");
		}
	});


	// resize docs

	var uprIsResizing = false;

	var upr = $("#util_pane_right");

	upr.on('mousedown', function(e) {

		if (e.pageX >= upr.offset().left + upr.width() - 3) {
			uprIsResizing = true;
		}
	});

	upr.on('mousemove', function(e) {
		if (e.pageX >= upr.offset().left + upr.width() - 3) {
			upr.css('cursor', 'col-resize');
		} else {
			upr.css('cursor', 'default');
		}
	});

	$(document).on('mousemove', function(e) {
		if (!uprIsResizing)
			return;

		var offsetRight = + e.clientX - upr.offset().left;

		upr.css('flex', '0 0 ' + offsetRight + 'px');
		$("#doc_list").jqxListBox({ width: '100%'});
		$("#jumps_list").jqxListBox({ width: '100%'});
	});



	$(document).on('mouseup', function(e) {
		if (uprIsResizing) {
			localStorage['VIEWPORT_DOCUMENT_PANE_SIZE'] = e.clientX - upr.offset().left;
			_editor.refresh();
		}
		uprIsResizing = false;
	});



	// resize projects

	var projIsResizing = false;

	var proj = $("#util_pane_left");

	proj.on('mousedown', function(e) {

		if (e.pageX >= proj.offset().left + proj.width() - 5) {
			projIsResizing = true;
		}

	});

	proj.on('mousemove', function(e) {
		if (e.pageX >= proj.offset().left + proj.width() - 5) {
			proj.css('cursor', 'col-resize');
		} else {
			proj.css('cursor', 'default');
		}
	});

	$(document).on('mousemove', function(e) {
		if (!projIsResizing)
			return;

		var offsetRight = + e.clientX - proj.offset().left;

		proj.css('flex', '0 0 ' + offsetRight + 'px');
		// $("#doc_list").jqxListBox({ width: '100%'});
		// $("#jumps_list").jqxListBox({ width: '100%'});
	});


	$(document).on('mouseup', function(e) {
		if (projIsResizing) {
			var w = e.clientX - proj.offset().left;
			if (w <= 3) {
				w = 3;
				proj.css('flex', '0 0 3px');
			}
			localStorage['PROJECTPANE_WIDTH'] = w;
			_editor.refresh();
		}
		projIsResizing = false;
	});



	function setUiZoom() {


		var html = 'Set UI zoom (100 is 100%)<br><input id="zoom_ui_input" autofocus style="font-size: .9em; width: 95%; outline: none; border: none; background-color: transparent;"></input> <div style="font-size: 1em; float: right;">Hit enter to confirm!</div>';


		_editor.openDialog(html, function(value) {

			value = isNaN(value) ? 100 : value;

			$("html").css({"font-size": value.toString() + "%"});
			redrawListDirty();

			localStorage['VIEWPORT_UI_ZOOM'] = value;

		}, {
			value: localStorage['VIEWPORT_UI_ZOOM'] ? localStorage['VIEWPORT_UI_ZOOM'] : '100',
			selectValueOnOpen: true,
			closeOnEnter: true,
			closeOnBlur: false,
			onClose: function() {}
		});
	}




	/*
	 ######     #######   ##     ##  ##    ##   ########    ######
	##    ##   ##     ##  ##     ##  ###   ##   ##     ##  ##    ##
	##         ##     ##  ##     ##  ####  ##   ##     ##  ##
	 ######    ##     ##  ##     ##  ## ## ##   ##     ##   ######
	      ##   ##     ##  ##     ##  ##  ####   ##     ##        ##
	##    ##   ##     ##  ##     ##  ##   ###   ##     ##  ##    ##
	 ######     #######    #######   ##    ##   ########    ######
	*/

	function getWhiteNoiseSounds() {
		var filesA = fs.readdirSync("sounds/white_noise/");

		// construct palleteItems
		var objectsA = [];

		objectsA.push({
			string: "Stop All Sounds",
			shortcut: "",
			action: 'wm.stopWhiteNoise',
			type: "exe_global",
			sublist: [],
			icon: "&#xf011;",
			persist: true
		});

		for (var i = 0; i < filesA.length; i++) {



			var item = {
				string: "Play: " + filesA[i].replace(/\..*$/, ""),
				shortcut: "",
				action: 'wm.playWhiteNoise("' + filesA[i] + '")',
				type: "eval",
				sublist: [],
				icon: "&#xf025;",
				persist: true
			};
			objectsA.push(item);
		}

		return objectsA;
	}


	var whiteNoiseSound;
	function playWhiteNoise(fileName) {
		localStorage["SOUNDS_WHITE_NOISE_SELECTED_SCHEME"] = fileName.replace(/\..+/, "");
		if (whiteNoiseSound) whiteNoiseSound.fade(0.7, 0, 1000);

		whiteNoiseSound = new Howl({
			urls: ['./sounds/white_noise/' + fileName],
			autoplay: false,
			loop: true,
			volume: 0
		});
		whiteNoiseSound.play().fade(0, 0.7, 1000);
		showNotice("Playing: " + fileName.replace(/\..*$/, ""));
	}

	function stopWhiteNoise() {
		if (whiteNoiseSound)
			whiteNoiseSound.fade(0.7, 0, 1000, function() {
				whiteNoiseSound.unload();
				localStorage["SOUNDS_WHITE_NOISE_SELECTED_SCHEME"] = "";
			});
	}

	/*
	########   ########   ##     ##  ########   ##            ###     ########   ########    ######
	   ##      ##         ###   ###  ##     ##  ##           ## ##       ##      ##         ##    ##
	   ##      ##         #### ####  ##     ##  ##          ##   ##      ##      ##         ##
	   ##      ######     ## ### ##  ########   ##         ##     ##     ##      ######      ######
	   ##      ##         ##     ##  ##         ##         #########     ##      ##               ##
	   ##      ##         ##     ##  ##         ##         ##     ##     ##      ##         ##    ##
	   ##      ########   ##     ##  ##         ########   ##     ##     ##      ########    ######
	*/


	function getCssFromTemplateFile(name) {
		var templateString = "";
		var templatePath = "templates/" + name;

		try {
			templateString = fs.readFileSync(templatePath, {
				encoding: "utf-8"
			});
		} catch (e) {
			return "";
		}

		return templateString;
	}


	function getJsFromTemplateFile(name) {
		var templateString = "";
		var templatePath = "templates/" + name.replace(/\..*$/, "") + ".js";
		var defaultTemplatePath = "templates/_default.js";

		try {
			templateString = fs.readFileSync(templatePath, {
				encoding: "utf-8"
			});
		} catch (e) {
			try {
				templateString = fs.readFileSync(defaultTemplatePath, {
					encoding: "utf-8"
				});

			} catch (ex) {
				return "";
			}
		}

		return templateString;
	}


	function getInstalledTemplateNames() {
		return fs.readdirSync("templates/");
	}

	function setTemplate(name) {
		localStorage["EXPORT_TEMPLATE"] = name;
	}




	/*
	########   ##     ##  ########   ##     ##  ########    ######
	   ##      ##     ##  ##         ###   ###  ##         ##    ##
	   ##      ##     ##  ##         #### ####  ##         ##
	   ##      #########  ######     ## ### ##  ######      ######
	   ##      ##     ##  ##         ##     ##  ##               ##
	   ##      ##     ##  ##         ##     ##  ##         ##    ##
	   ##      ##     ##  ########   ##     ##  ########    ######
	*/
	function getThemes() {

		var objectsA = [];
		var filesA = fs.readdirSync("themes/");



		for (var i = 0; i < filesA.length; i++) {

			// exclude non css files
			if (!filesA[i].endsWith(".css")) continue;

			var item = {
				string: "Theme: " + filesA[i].replace(/\..*$/, ""),
				shortcut: "",
				action: 'wm.setTheme("' + filesA[i] + '")',
				type: "eval",
				sublist: [],
				icon: "&#xf1fc;",
				persist: true,
				widget: null
			};
			objectsA.push(item);
		}


		return objectsA;

	}


	function getFonts() {

		var objectsA = [];
		var fontsA = _settings.main_font_list;

		for (var i = 0; i < fontsA.length; i++) {

			var item = {
				string: 'Set font: ' + fontsA[i],
				shortcut: "",
				action: 'wm.setFont("' + fontsA[i] + '");',
				type: "eval",
				sublist: [],
				icon: "&#xf031;",
				persist: true,
				widget: null
			};

			objectsA.push(item);
		}

		// also set line spacing items
		var lineSpacingA = _settings.main_font_line_spacing_list;

		for (var ii = 0; ii < lineSpacingA.length; ii++) {

			var item = {
				string: "Line spacing to: " + lineSpacingA[ii],
				shortcut: "",
				action: 'wm.setLineSpacing(' + lineSpacingA[ii] + ');',
				type: "eval",
				sublist: [],
				icon: "&#xf0c9;",
				persist: true,
				widget: null
			};

			objectsA.push(item);
		}


		return objectsA;
	}


	var themeTranslateItems = [{
		wm: "$wm-ui-general",
		css: "body",
		prp: false
	}, {
		wm: "$wm-ui-editor-viewport",
		css: "#main_pane",
		prp: false
	}, {
		wm: "$wm-ui-writing-area",
		css: "#editor-placeholder",
		prp: false
	}, {
		wm: "$wm-ui-list-selected",
		css: ".jqx-fill-state-pressed-metro",
		prp: false
	}, {
		wm: "$wm-ui-list-hover",
		css: ".jqx-fill-state-hover-metro",
		prp: false
	}, {
		wm: "$wm-ui-list-input",
		css: ".jqx-listbox-filter-input-metro, #module_pane ::-webkit-input-placeholder, #module_pane input, #sources_list input",
		prp: false
	}, {
		wm: "$wm-ui-module-pane",
		css: "#module_pane",
		prp: false
	}, {
		wm: "$wm-ui-project-pane",
		css: "#util_pane_left, .jqx-tree-item-metro",
		prp: false
	}, {
		wm: "$wm-ui-document-pane",
		css: "#doc_list, #jumps_list, #doc_header, #doc_filter, #doc_search", // .jqx-listitem-element span
		prp: false
	}, {
		wm: "$wm-ui-icons",
		css: "#doc_header, #doc_filter, #module_selector, .CodeMirror-guttermarker-subtle, .jqx-tree-item-arrow-expand-metro, .jqx-tree-item-arrow-collapse-metro",
		prp: false
	}, {
		wm: "$wm-ui-list-divider",
		css: ".jqx-listitem-element span",
		prp: false
	}, {
		wm: "$wm-repo-tile",
		css: ".wm-repo-tile",
		prp: false
	}, {
		wm: "$wm-repo-clickable",
		css: ".wm-repo-clickable",
		prp: false
	}, {
		wm: "$wm-repo-comment",
		css: ".wm-repo-comment",
		prp: false
	}, {
		wm: "$wm-ui-infobar",
		css: "#info_bar_bottom, #info_bar_top",
		prp: false
	}, {
		wm: "$wm-editor",
		css: ".CodeMirror",
		prp: true
	}, {
		wm: "$wm-formatting",
		//css: " .cm-formatting:not(.cm-meta):not(.cm-property):not(.cm-variable-2):not(.cm-variable-1):not(.cm-variable-3):not(.cm-keyword)",
		css: " .cm-formatting",
		prp: true
	}, {
		wm: "$wm-selection",
		css: " .CodeMirror-selected, .CodeMirror-focused .CodeMirror-selected, ::selection:not(.searchFilter), .CodeMirror ::selection", // .CodeMirror-selectedtext
		prp: true
	}, {
		wm: "$wm-active-line",
		css: " .CodeMirror-activeline-background",
		prp: true
	}, {
		wm: "$wm-active-sentence",
		css: " .CodeMirror-activesentence-background",
		prp: true
	},{
		wm: "$wm-cursor",
		css: " .CodeMirror-cursor",
		prp: true
	}, {
		wm: "$wm-comment",
		css: " .cm-wm-comments",
		prp: true
	}, {
		wm: "$wm-marked",
		css: " .cm-wm-marked",
		prp: true
	}, {
		wm: "$wm-footnote",
		css: " .cm-wm-footnote",
		prp: true
	}, {
		wm: "$wm-anchored-reading",
		css: " .cm-wm-anchored-reading",
		prp: true
	}, {
		wm: "$wm-deleted",
		css: " .cm-strikethrough",
		prp: true
	}, {
		wm: "$wm-whitespace-redundantspace",
		css: ".cm-wm-rspace",
		prp: false
	}, {
		wm: "$wm-whitespace-tab",
		css: ".cm-tab:before",
		prp: false
	}, {
		wm: "$wm-whitespace-newline",
		css: "pre.CodeMirror-line > span:after",
		prp: false
	}, {
		wm: "$wm-spell-error",
		css: " .cm-spell-error:not(.cm-url):not(.cm-link):not(.cm-wm-comments):not(.cm-wm-metadata):not(.cm-strikethrough):not(.cm-comment)",
		prp: true
	}, {
		wm: "$wm-task-unchecked",
		css: " span.cm-meta",
		prp: true
	}, {
		wm: "$wm-task-checked",
		css: " span.cm-property",
		prp: true
	}, {
		wm: "$wm-search-match",
		css: " .cm-searching",
		prp: true
	}, {
		wm: "$wm-scrollbar-matches",
		css: ".CodeMirror-search-match",
		prp: false
	}, {
		wm: "$wm-metadata",
		css: " span.cm-wm-metadata",
		prp: true
	}, {
		wm: "$wm-bold",
		css: " span.cm-strong",
		prp: true
	}, {
		wm: "$wm-italic",
		css: " span.cm-em",
		prp: true
	}, {
		wm: "$wm-break",
		css: " .cm-hr:after",
		prp: true
	}, {
		wm: "$wm-headings",
		css: " .cm-header",
		prp: true
	}, {
		wm: "$wm-heading-1",
		css: " .cm-header-1",
		prp: true
	}, {
		wm: "$wm-heading-2",
		css: " .cm-header-2",
		prp: true
	}, {
		wm: "$wm-heading-3",
		css: " .cm-header-3",
		prp: true
	}, {
		wm: "$wm-heading-4",
		css: " .cm-header-4",
		prp: true
	}, {
		wm: "$wm-heading-5",
		css: " .cm-header-5",
		prp: true
	}, {
		wm: "$wm-heading-6",
		css: " .cm-header-6",
		prp: true
	}, {
		wm: "$wm-code",
		css: " span.cm-comment",
		prp: true
	}, {
		wm: "$wm-list-item-1",
		css: " span.cm-variable-2",
		prp: true
	}, {
		wm: "$wm-list-item-2",
		css: " span.cm-variable-3",
		prp: true
	}, {
		wm: "$wm-list-item-3",
		css: " span.cm-keyword",
		prp: true
	}, {
		wm: "$wm-quotes",
		css: " .cm-quote",
		prp: true
	}, {
		wm: "$wm-link",
		css: " .cm-link",
		prp: true
	}, {
		wm: "$wm-url",
		css: " .cm-quote",
		prp: true
	}, {
		wm: "$wm-image",
		css: " span.cm-tag",
		prp: true
	}, {
		wm: "$wm-notice",
		css: ".wm-notice, .jqx-tooltip, .jqx-tooltip-text, .jqx-tooltip-main",
		prp: false
	}, {
		wm: "$wm-collapsed-placeholder",
		css: ".collapsed-placeholder",
		prp: false
	}, {
		wm: "$wm-ui-progress-background",
		css: "#info_bar_progressbar_outer",
		prp: false
	}, {
		wm: "$wm-ui-progress-bar",
		css: "#info_bar_progressbar_inner",
		prp: false
	}, {
		wm: "$wm-ui-progress-session",
		css: "#info_bar_progressbar_session",
		prp: false
	}];



	function translateThemeFile(cssString, themeName) {
		var result = parseThemeVariables(cssString);
		for (var i = 0; i < themeTranslateItems.length; i++) {
			var prp = (themeTranslateItems[i].prp) ? ".cm-s-" + themeName : "";
			result = result.replace(themeTranslateItems[i].wm,  '.' + themeName + ' ' + prp + themeTranslateItems[i].css);
		}
		return result;
	}


	function parseThemeVariables(s) {
		var st = s;

		var varsBlock = st.match(/\:root\s*{((.|\s)+?)}/);
		if (varsBlock && varsBlock[1]) { // variables found


			varsBlock[1] = varsBlock[1].replace(/\/\*[^*]*\*+([^/*][^*]*\*+)*\//g, ""); // remove comments
			var lines = varsBlock[1].trim().split(";");

			// remove blanks
			lines = lines.filter(function(val) { return val.trim() !== ""; });

			// get array with key value pairs
			var pairs = lines.map(function(item) {
				var arr = item.trim();
				var pair = arr.split(":");
				var key = pair[0].trim();
				var value = pair[1].trim();
				return {key: key, value: value};
			});


			// replace var(varname) with values

			var newCSS = st;
			for (var i = 0; i < pairs.length; i++) {
				var r = new RegExp('var\\(' + pairs[i].key + '\\)', 'g');
				newCSS = newCSS.replace(r, pairs[i].value);
			}

			return newCSS; // return new parsed string

		} else {
			return st; // no variables found, return original string
		}
	}



	function setTheme(theme) {


		var stored = (localStorage["VIEWPORT_THEME"]) ? localStorage["VIEWPORT_THEME"] : "wm-default.css";
		var theme = (theme) ? theme : stored;
		var themePath = "themes/" + theme;
		var themeName = theme.replace(/\..*$/, "");


		// apply fallback theme from core
		if (theme == "fallback") {
			themePath = "core/fallback-theme.css";
			themeName = "fallback-theme";
		}


		// Set codemirror option
		_editor.setOption("theme", themeName);

		// add controll class to html
		if (_currentTheme != "") { $("html").removeClass(_currentTheme); }
		$("html").addClass(themeName);
		_currentTheme = themeName;

		fs.readFile(themePath, function(err, data) {
			if (err) {
				setTheme("fallback");
				return;
			}

			var css = translateThemeFile(data.toString(), themeName);

			// preprocess existing file
			var head = document.getElementsByTagName('head')[0];
			var style = document.createElement("style");
			style.appendChild(document.createTextNode(css));
			head.appendChild(style);
			localStorage["VIEWPORT_THEME"] = theme;


			// store basic colors
			localStorage["VIEWPORT_TEXT_COLOR"] = $(".CodeMirror").css("color");
			localStorage["VIEWPORT_BACKGROUND_COLOR"] = $("body").css("background-color");
			localStorage["VIEWPORT_MARK_COLOR"] = $("#doc_header").css("color");

			// set base universal classes
			setBaseColorClasses(themeName, localStorage["VIEWPORT_TEXT_COLOR"], localStorage["VIEWPORT_BACKGROUND_COLOR"], localStorage["VIEWPORT_MARK_COLOR"]);


			// hacks
			$("head").append("<style>" + '.' + themeName + " .cm-wm-sentence-breakup::after{ border-bottom-color:" + localStorage["VIEWPORT_MARK_COLOR"] + "; border-bottom-style: " + _settings.sentence_breakup_divider_style + ";}</style>");


			// set custom scrollbar colors
			$("head").append("<style>" + '.' + themeName + " .jqx-scrollbar-thumb-state-normal-metro, .CodeMirror-overlayscroll-vertical div, .ps-scrollbar-y { background-color: " + localStorage["VIEWPORT_MARK_COLOR"] + "!important; opacity: .3; }</style>");

			// change font if set
			if (localStorage["VIEWPORT_MAIN_FONT"] && localStorage["VIEWPORT_MAIN_FONT"] != "") {
				setFont(localStorage["VIEWPORT_MAIN_FONT"]);
			} else {
				$(".CodeMirror").css('font-family', 'Hack Mono'); // default
			}

			// set line spacing if set
			if (localStorage["VIEWPORT_MAIN_FONT_LINE_SPACING"] && localStorage["VIEWPORT_MAIN_FONT_LINE_SPACING"] != "") {
				setLineSpacing(localStorage["VIEWPORT_MAIN_FONT_LINE_SPACING"]);
			} else {
				$(".CodeMirror").css('line-height', '1.5em'); // default
			}

			_editor.refresh();

		});
	}


	function setFont(font) {
		$(".CodeMirror").css('font-family', font);
		localStorage["VIEWPORT_MAIN_FONT"] = font;
		_editor.refresh();
	}


	function setLineSpacing(spacing) {
		var spacingPlus = .35 + parseFloat(spacing);
		$(".CodeMirror").css('line-height', spacingPlus.toString() + "em");

		localStorage["VIEWPORT_MAIN_FONT_LINE_SPACING"] = spacing;
		_editor.refresh();
	}



	/*
	 ######     #######   ##          #######   ########    ######
	##    ##   ##     ##  ##         ##     ##  ##     ##  ##    ##
	##         ##     ##  ##         ##     ##  ##     ##  ##
	##         ##     ##  ##         ##     ##  ########    ######
	##         ##     ##  ##         ##     ##  ##   ##          ##
	##    ##   ##     ##  ##         ##     ##  ##    ##   ##    ##
	 ######     #######   ########    #######   ##     ##   ######
	*/

	function setBaseColorClasses(themeName, textColor, backColor, markColor) {

		var brightness = tinycolor(markColor).getBrightness();
		var markModifiedColor = (brightness > 50 ? tinycolor(markColor).darken(5) : tinycolor(markColor).lighten(15)).toString();

		setBaseColorClass(themeName, "theme-text-color", "color", textColor);
		setBaseColorClass(themeName, "theme-background-color", "background-color", backColor);
		setBaseColorClass(themeName, "theme-mark-color", "color", markColor);

		setBaseColorClass(themeName, "theme-text-color-inverse", "color", backColor);
		setBaseColorClass(themeName, "theme-background-color-inverse", "background-color", textColor);
		setBaseColorClass(themeName, "theme-background-color-mark", "background-color", markColor);


		setBaseColorClass(themeName, "theme-background-color-mark-modified", "background-color", markModifiedColor);


		setBaseColorClass(themeName, "theme-border-color", "border-color", textColor);
		setBaseColorClass(themeName, "theme-border-color-mark", "border-color", markColor);
		setBaseColorClass(themeName, "theme-border-color-inverse", "border-color", backColor);
		setBaseColorClass(themeName, "theme-border-color-mark-modified", "border-color", markModifiedColor);


		// hacks
		setBaseColorClass(themeName, "jqx-tree-item", "background-color", "rgba(0,0,0,0)");

		setBaseColorClass(themeName, "palette-results::-webkit-scrollbar-thumb", "background-color", markModifiedColor);

	}

	function setBaseColorClass(themeName, className, propertyName, propertyValue) {
		$("head").append('<style> .' + themeName + ' .' + className + ' { ' + propertyName + ': ' + propertyValue + ';}</style>');
	}


	/*
	########   ####  ##         ########   ########   ########    ######
	##          ##   ##            ##      ##         ##     ##  ##    ##
	##          ##   ##            ##      ##         ##     ##  ##
	######      ##   ##            ##      ######     ########    ######
	##          ##   ##            ##      ##         ##   ##          ##
	##          ##   ##            ##      ##         ##    ##   ##    ##
	##         ####  ########      ##      ########   ##     ##   ######
	*/

	function getFilterSchemes() {
		var objectsA = [];
		var schemes = _settings.ui_filter_schemes;


		for (var i = 0; i < schemes.length; i++) {
			var item = {
				string: "Filter: " + schemes[i].name,
				shortcut: "",
				icon: "&#xf042;",
				action: 'wm.applyFilters("' + filterStringFromObject(schemes[i].filters) + '")',
				type: "eval",
				sublist: [],
				persist: true
			};
			objectsA.push(item);
		}
		return objectsA;
	}


	function filterStringFromObject(obj) {
		var filterString = "";
		for (var key in obj) {
			if (obj.hasOwnProperty(key)) {
				filterString += key + "(" + obj[key] + ") ";
			}
		}
		return filterString;
	}

	function objectFromFilterString(str) {
		var obj = {};
		var split1 = str.trim().split(/\s+/);
		if (split1.length < 1) split1 = [str];

		for (var i = 0; i < split1.length; i++) {
			var split2 = split1[i].trim().split("(");
			var key = split2[0].trim();
			var value = split2[1].replace(")", "").trim();
			obj[key] = value;
		}
		return obj;
	}


	function applyFilters(str) {
		var storedFilters = (localStorage["VIEWPORT_FILTERS"]) ? localStorage["VIEWPORT_FILTERS"] : "none";
		var str = (str) ? str : storedFilters;
		localStorage["VIEWPORT_FILTERS"] = str;
		$("body").css("-webkit-filter", "none");
		$("body").css("-webkit-filter", str);
	}


	function resetAllFilters() {
		localStorage["VIEWPORT_FILTERS"] = "none";
		$("body").css("-webkit-filter", "none");
	}



	function rotateBrightness() {
		var val;
		var storedFilters = (localStorage["VIEWPORT_FILTERS"] && localStorage["VIEWPORT_FILTERS"] != "none") ? localStorage["VIEWPORT_FILTERS"] : "brightness(100%)";
		var obj = objectFromFilterString(storedFilters);

		val = (obj["brightness"]) ? parseInt(obj["brightness"].replace("%", ""), 10) : 100;
		val -= 5;
		val = (val < 20) ? 100 : val;
		obj["brightness"] = "" + val + "%";

		applyFilters(filterStringFromObject(obj));

	}


	function rotateSepia() {
		var val;
		var storedFilters = (localStorage["VIEWPORT_FILTERS"] && localStorage["VIEWPORT_FILTERS"] != "none") ? localStorage["VIEWPORT_FILTERS"] : "sepia(0%)";
		var obj = objectFromFilterString(storedFilters);

		val = (obj["sepia"]) ? parseInt(obj["sepia"].replace("%", ""), 10) : 0;
		val += 7;
		val = (val > 70) ? 0 : val;
		obj["sepia"] = "" + val + "%";

		applyFilters(filterStringFromObject(obj));

	}



	/*
	##     ##  ####       ########   ##     ##  ########   ##    ##   ########    ######
	##     ##   ##        ##         ##     ##  ##         ###   ##      ##      ##    ##
	##     ##   ##        ##         ##     ##  ##         ####  ##      ##      ##
	##     ##   ##        ######     ##     ##  ######     ## ## ##      ##       ######
	##     ##   ##        ##          ##   ##   ##         ##  ####      ##            ##
	##     ##   ##        ##           ## ##    ##         ##   ###      ##      ##    ##
	 #######   ####       ########      ###     ########   ##    ##      ##       ######
	*/



	$("#info_bar_top").on("dblclick", function(e) {
		_editor.execCommand("goDocStart");
		_editor.execCommand("showScrollBar");
	});


	$("#info_bar_bottom").on("dblclick", function(e) {
		_editor.execCommand("goDocEnd");
		_editor.execCommand("showScrollBar");
	});




	function showProjectWordCount() {

		if (_currentProjectId == -1) return;

		storage.getCombinedProject(_currentProjectId, function(textsA) {

			var txt = textsA.map(function(item) {
				return stats.removeMetaPart(item.txt).trim();
			}).join('\n\n');

			var words = stats.getWords(txt).count;
			var chars = stats.getCharactersWS(txt).count;
			// improvement
			// get the same stat as shown in Infobr
			//var words = eval('stats.' + localStorage['STAT_IN_INFOBAR'] + '(txt).count');

			showNotice('<div><strong style="font-size: 1.5em;">' + stats.add000separator(words.toString()) + '</strong> words<br><strong style="font-size: 1.5em;">' + stats.add000separator(chars.toString()) + '</strong> characters<br><br>in ' + textsA.length + ' documents</div>', {
				autoClose: true,
				autoCloseDelay: 10000,
				width: 13
			})

		});
	}



	$("#info_bar_right").on("contextmenu click", function(e) {
		menus.getInfobarStatsMenu().popup(e.originalEvent.clientX, e.originalEvent.clientY);
	});


	/*
	 * Set native OS menu bar
	 */

	var nativeMenuBar = new gui.Menu({
		type: 'menubar'
	});


	if (os.platform() === 'darwin') {
		nativeMenuBar.createMacBuiltin('Writemonkey', {
			hideEdit: false,
			hideWindow: false
		});
		win.menu = nativeMenuBar;
	} else if (os.platform() === 'linux') {

	} else if (os.platform() === 'win32') {

	}



	/*
	########   ########   ####  ##    ##   ########
	##     ##  ##     ##   ##   ###   ##      ##
	##     ##  ##     ##   ##   ####  ##      ##
	########   ########    ##   ## ## ##      ##
	##         ##   ##     ##   ##  ####      ##
	##         ##    ##    ##   ##   ###      ##
	##         ##     ##  ####  ##    ##      ##
	*/

	function printText(txt) {

		var text = (_editor.somethingSelected()) ? _editor.getSelection("\n\n") : _editor.getValue();
		text = (txt) ? txt : text;

		var textToPrint =  text.replace(/(?:\r\n|\r|\n)/g, '<br>').replace(/\t/g, '&emsp;');

		$ifr = $('<iframe id="print-iframe" style="background-color: transparent; position: absolute; width: 0px; height: 0px; visibility: hidden;" nwfaketop nwdisable></iframe>').appendTo("body");


		var doc = $ifr[0].document;
		if ($ifr[0].contentDocument) {
			doc = $ifr[0].contentDocument;
		} else if ($ifr[0].contentWindow) {
			doc = $ifr[0].contentWindow.document;
		}

		doc.open();
		doc.writeln('<html><head><title>' + _currentDoc.nam + '</title><style type="text/css"> body {font-size: ' + _settings.plain_text_print_font_size + '; font-family: '+ _settings.plain_text_print_font + ';}</style></head><body>' + textToPrint + '</body></html>');
		doc.close();
		$ifr[0].contentWindow.print();

		$($ifr).remove();
	}





	/*
	##         ####   ######    ########   ##    ##    ######    ########
	##          ##   ##    ##   ##         ###   ##   ##    ##   ##
	##          ##   ##         ##         ####  ##   ##         ##
	##          ##   ##         ######     ## ## ##    ######    ######
	##          ##   ##         ##         ##  ####         ##   ##
	##          ##   ##    ##   ##         ##   ###   ##    ##   ##
	########   ####   ######    ########   ##    ##    ######    ########
	*/

	function storeKey(k) {
		// check if key is valid
		var name = ckk(k);
		if (name == -1 || k.length <= 0) {
			showNotice("Provided license key is not valid!", {autoClose: true, autoCloseDelay: 5000});
			localStorage.removeItem(atob("TElDRU5TRV9LRVk="));

		} else {
			localStorage[atob("TElDRU5TRV9LRVk=")] = k;
			showNotice('Thanks ' + name + '. Your key is valid! Restart Writemonkey!', {autoClose: false, autoCloseDelay: 5000});
		}

	}


	function enterWmLicenseKey() {
		var shortcut = (os.platform() === 'darwin') ? 'CMD+V' :  'CTRL+V';
		var html = 'Paste ('+ shortcut +') license key here: <br><input autofocus style="font-size: .9em; width: 95%; outline: none; border: none; background-color: transparent;"></input> <div style="font-size: 1em; float: right;">Hit enter to confirm!</div>';

		//_editor.openConfirm(html, [function(value) {storeKey(value);}, function(value) {storeKey(value);}], {value: "k jaslkdj lkasdj"});

		_editor.openDialog(html, function(value) {storeKey(value)}, {
			value: localStorage[atob("TElDRU5TRV9LRVk=")] && localStorage[atob("TElDRU5TRV9LRVk=")].length > 0 ? localStorage[atob("TElDRU5TRV9LRVk=")] : "",
			selectValueOnOpen: true,
			closeOnEnter: true,
			closeOnBlur: true,
			onClose: function() {}
		});
	}

	function removeWmLicenseKey() {
		localStorage.removeItem(atob("TElDRU5TRV9LRVk="));
		showNotice('The license was removed. After the restart the premium features won\'t work!', {autoClose: false, autoCloseDelay: 5000});
	}



	// new version: var d = crp.createDecipheriv( 'aes-256-ctr', pw, iv);
	function dcrptk(s) {
		var d = crp.createDecipher('aes-256-ctr', 'Arial Narrow');
		try {
			var de = d.update(s, 'hex', 'utf8');
		} catch (e) {
			return "";
		}
		de += d.final('utf8');
		return de;
	}




	function getLk() {
		var k = atob("TElDRU5TRV9LRVk=");
		return localStorage[k] ? localStorage[k] : "";
	}


	function ckk(license) {
		var l = license ? license : getLk();
		var d = dcrptk(l);
		var s = d.split('=');
		if (s.length == 5 && s[3] == 'r' && s[4][1] == 'a') {
			return s[0];
		} else {
			localStorage.removeItem(atob("TElDRU5TRV9LRVk="));
			return -1;
		}
	}



	//window.wm = this;
	__init();


	/*
	   ###     ########   ####
	  ## ##    ##     ##   ##
	 ##   ##   ##     ##   ##
	##     ##  ########    ##
	#########  ##          ##
	##     ##  ##          ##
	##     ##  ##         ####
	*/
	return {
		enterWmLicenseKey: enterWmLicenseKey,
		removeWmLicenseKey: removeWmLicenseKey,
		ckk: ckk,
		toggleDistractionFree: toggleDistractionFree,
		toggleAutoHide: toggleAutoHide,
		hidePanes: hidePanes,
		showPanes: showPanes,
		setUiZoom: setUiZoom,
		isWmFullScreen: isWmFullScreen,
		toggleDocsJumps: toggleDocsJumps,
		toggleLockOnTop: toggleLockOnTop,
		togglePanes: togglePanes,
		showNotice: showNotice,
		updateIconBadge: updateIconBadge,
		getWhiteNoiseSounds: getWhiteNoiseSounds,
		playWhiteNoise: playWhiteNoise,
		stopWhiteNoise: stopWhiteNoise,
		filterStringFromObject: filterStringFromObject,
		getFilterSchemes: getFilterSchemes,
		applyFilters: applyFilters,
		resetAllFilters: resetAllFilters,
		rotateBrightness: rotateBrightness,
		rotateSepia: rotateSepia,
		getThemes: getThemes,
		getFonts: getFonts,
		setFont: setFont,
		setLineSpacing: setLineSpacing,
		setTheme: setTheme,
		getInstalledTemplateNames: getInstalledTemplateNames,
		setTemplate: setTemplate,
		getCssFromTemplateFile: getCssFromTemplateFile,
		getJsFromTemplateFile: getJsFromTemplateFile,
		animateProgressBar: animateProgressBar,
		startNewSession: startNewSession,
		messageToStatsWorker: messageToStatsWorker,
		pauseUnpauseCurrentSession: pauseUnpauseCurrentSession,
		stopCurrentSession: stopCurrentSession,
		printText: printText,
		showProjectWordCount: showProjectWordCount
	};

})(window);