module.exports = function (api, pluginInfo) {

	var monkey = window;
	var e = window._editor;
	var options = pluginInfo.options;
	var $ = api.window.$;


	var currentCommand = "";

	function _init() {
		setPaletteItems();
	}

	function setPaletteItems() {

		monkey.commandPalette.registerPaletteItem({
			id: 'pandoc-set-palette',
			string: 'Pandoc Export/Import: Select Action',
			shortcut: "",
			action: '',
			type: "sublist",
			sublist: '_plugins["pandoc"].getPandocPaletteCommands()',
			persist: false,
			icon: "&#xf074;",
			widget: {
				type: "eval",
				value: '"Pandoc must be installed separately"'
			}
		});

	}

	function getPandocPaletteCommands() {

		var objectsA = [];
		var pandocCommands = options.pandocCommands;

		for (var i = 0; i < pandocCommands.length; i++) {

			var item = {
				string: pandocCommands[i].name,
				shortcut: "",
				action: '_plugins["pandoc"].runPandocCommand("' + i.toString() + '")',
				type: "eval",
				sublist: [],
				icon: pandocCommands[i].type == "export" ? "&#xf08b;" : "&#xf090;",
				persist: false,
				widget: {
					type: "eval",
					value: '" ' + pandocCommands[i].description + '  "'
				}
			};

			objectsA.push(item);
		}

		// put last to top
		if (monkey.localStorage["PANDOC_LAST_COMMAND"] && parseInt(monkey.localStorage["PANDOC_LAST_COMMAND"], 10) < objectsA.length) {
			objectsA.unshift(objectsA.splice(parseInt(monkey.localStorage["PANDOC_LAST_COMMAND"], 10), 1)[0]);
			objectsA[0].shortcut = "last used";
		}

		return objectsA;
	}

	// Entry


	function runPandocCommand(num) {

		monkey.localStorage["PANDOC_LAST_COMMAND"] = num;
		currentCommand = options.pandocCommands[parseInt(num, 10)];

		var args = replaceCommandVariables((currentCommand.attachGlobalOptions ? options.globalOptions + " " : "") + currentCommand.command);

		// also add args from pandoc meta field
		var meta = monkey.storage.getMetaFields();
		if (meta.pandoc && meta.pandoc != "") {
			args = meta.pandoc + " " + args;
		}
		args = args.replace(/\s{2,}/g, " "); // remove all double spaces

		if (currentCommand.source == "file") {
			// get file path from dialog and then run pandocExec
			var dir = monkey.files.getLastUsedPath();
			var hiddenInput = $('<input style="display:none;" id="fileDialogOpen" type="file" nwworkingdir="' + dir + '" />');
			$("body").append(hiddenInput);
			hiddenInput.unbind('change');

			hiddenInput.change(function (evt) {
				var src = $(this).val();

				$(this).val('');
				pandocExec(src, args);
			});

			hiddenInput.trigger('click');

		} else {
			var src = removeComments(getSource(currentCommand.source));
			pandocExec(src, args);
		}
	}

	function pandocExec(src, args) {
		showPandocNotice("Export started ...", currentCommand.name);
		monkey.nodePandoc(src, args, exportCallback);
	}

	function exportCallback(err, result, argsA) {
		if (err) {
			monkey.wm.showNotice(err.message.toString(), {
				autoClose: false,
			});
		} else {
			// success
			if (currentCommand.target == "clipboard") {
				console.log(result)
				copyToClipboard(result)
			} else if (result === true && currentCommand.target == "file") {
				// do nothing, just show notice
				showPandocNotice("Export to file successful!", currentCommand.description);
				if (currentCommand.openInApp) {
					openInApp(argsA)
				}
			}
		}
	}

	function getSource(sourceType) {
		var source;
		switch (sourceType) {
			case "editor":
				if (e.somethingSelected()) {
					source = e.getSelection();
				} else {
					source = e.getValue();
				}
				return source;
				break;

			case "clipboard":
				var clipboard = monkey.gui.Clipboard.get();
				return clipboard.get('text');
				break;

			default:
				return false;
		}
	}

	function replaceCommandVariables(command) { // {name}, {desktop}, {temp}, {user}, {datetime}, {date}, {template}

		var homeDir = (process.platform === "win32") ? process.env.HOMEPATH : process.env.HOME;
		return command
			.replace(/{name}/g, monkey.storage.getDocumentName().replace(/[ \t]/g, "_").replace(/[\\~#%&*{}/:<>?|".,]+/g, ""))
			.replace(/{temp}/g, monkey.os.tmpdir())
			.replace(/{desktop}/g, monkey.path.join(homeDir, "Desktop"))
			.replace(/{datetime}/g, monkey.moment(new Date().getTime()).format(options.dateTimeFormat));
	}



	function openInApp(argsA) {
		monkey.opener(argsA[argsA.length - 1]); // is path always last in the array ???????
	}

	function copyToClipboard(result) {
		var clipboard = monkey.gui.Clipboard.get();
		clipboard.set(result, 'text');
		showPandocNotice("Export to clipboard successful!", currentCommand.description);
	}

	function showPandocNotice(notice, description) {
		monkey.wm.showNotice("<strong>" + notice + "</strong><br><br>" + description, {
			autoClose: true,
			autoCloseDelay: 10000
		});
	}


	function removeComments(text) {
		text = text.replace(/[^:]\/{2}(.*?)\/{2}/gm, '');
		text = text.replace(/^\/{2}(.*?)$/gm, '');

		return text;
	}


	return {
		_init: _init,
		runPandocCommand: runPandocCommand,
		getPandocPaletteCommands: getPandocPaletteCommands
	};

};