module.exports = function(api, pluginInfo) {

	var options = pluginInfo.options;
	var $ = api.window.$;
	var m = window;

	
	var levelUpTimer;
	var	levelDownTimer;
	var currentLevel = 1;
	var maxLevel = 1;
	var barElement = (options.bar_position == "left") ? $("#edge") : $("#edge2");
	
	
	function _init() {

		setCommands();
		setupStaminaBar();
	}


	function setupStaminaBar() {
		// Enable at start up
		if (m.localStorage.getItem("STATS_STAMINA_BAR") !== null && m.localStorage["STATS_STAMINA_BAR"] == "1") {
			enableStaminaBar();
		}
	}

	
	function resetDownTimer() {
		m.clearInterval(levelDownTimer);
		levelDownTimer = m.setInterval(function() {
			currentLevel = currentLevel - 1 < 2 ? 2 : currentLevel - 1;
			resetUpTimer();
			setLevel();
		}, options.level_down_after_seconds * 1000);
	}


	function resetUpTimer() {
		m.clearInterval(levelUpTimer);
		levelUpTimer = m.setInterval(function() {
			currentLevel = currentLevel + 1 > 99 ? 99 : currentLevel + 1;
			setLevel();
		}, options.level_up_after_seconds * 1000);
	}


	function setCommands() {

		// Command palette
		m.commandPalette.registerPaletteItem({
			id: 'stamina_bar',
			string: 'Stamina Bar: Toggle on / off ',
			shortcut: "",
			action: '_plugins["stamina-bar"].toggleStaminaBar()',
			type: 'eval',
			sublist: [],
			persist: false,
			icon: "&#xf0e4;",
			widget: {
				type: "setting",
				value: "STATS_STAMINA_BAR"
			}
		});
	}



	function setLevel() {

		maxLevel = currentLevel >= maxLevel ? currentLevel : maxLevel;
				
		var percent = currentLevel.toString() + "%";
		var plus = (maxLevel + 0.5).toString() + "%";	

		var orientation = options.bar_orientation;
		var barColor = m.localStorage["VIEWPORT_MARK_COLOR"];
		var markColor = m.localStorage["VIEWPORT_MARK_COLOR"];

		var gradient = ("-webkit-linear-gradient(orientation, barColor percent, transparent percent, transparent maxLevel%, markColor maxLevel%, markColor plus, transparent plus)")
		.replace(/orientation/g, orientation)
		.replace(/maxLevel/g, maxLevel)
		.replace(/plus/g, plus)
		.replace(/barColor/g, barColor)
		.replace(/markColor/g, markColor)
		.replace(/percent/g, percent);

		barElement.css("background", gradient);
	}





	function toggleStaminaBar() {
		
		if (m.localStorage["STATS_STAMINA_BAR"] == "1") {
			disableStaminaBar();
			currentLevel = 0;
			m.wm.showNotice("Stamina bar disabled.", {});

		} else {
			enableStaminaBar();
			m.wm.showNotice("Stamina bar enabled.\nMore you write, the higher it gets.\n\nETA for the zone is " + (options.level_up_after_seconds * 100 / 60).toFixed(1) + " minutes of uninterrupted writing.", {});
		}
	}

	function enableStaminaBar() {
		m._editor.on('change', resetDownTimer);
		resetUpTimer();
		resetDownTimer();
		m.localStorage["STATS_STAMINA_BAR"] = "1";
	}

	function disableStaminaBar() {
		maxLevel = 1;
		currentLevel = 1;
		barElement.css("background", "transparent");
		m.clearInterval(levelUpTimer);
		m.clearInterval(levelDownTimer);
		m._editor.off('change', resetDownTimer);
		m.localStorage["STATS_STAMINA_BAR"] = "0";
		m._editor.setOption("blastCode", false);
	}


	return {
		_init: _init,
		toggleStaminaBar: toggleStaminaBar
	};
};